import pytest
import copy
import sys

from lasagna.agent_util import (
    bind_model,
    build_simple_agent,
    flat_messages,
    noop_callback,
)

from lasagna.mock_provider import (
    MockProvider,
)

from lasagna.types import (
    AgentCallable,
    BoundAgentCallable,
    Message,
    ModelSpec,
    ToolResult,
    Model,
    EventCallback,
    AgentRun,
    AgentRunExtraction,
    MessageContent,
)

from typing import (
    List, Dict, Set, Callable, Awaitable,
    Literal, Union, Iterable, Any,
)

from enum import Enum

from lasagna.tools_util import (
    convert_to_json_schema,
    get_tool_params,
    handle_tools,
    build_tool_response_message,
    is_async_callable,
    is_callable_of_type,
    extract_tool_result_as_sting,
    validate_args,
    type_a_isa_b,
)


def test_convert_to_json_schema():
    j = convert_to_json_schema([
        {
            'name': 'a',
            'type': 'str',
            'description': 'param a',
        },
        {
            'name': 'second',
            'type': 'float',
            'description': 'second param',
        },
        {
            'name': 'another',
            'type': 'int',
            'description': 'another param',
        },
        {
            'name': 'yet_another',
            'type': 'bool',
        },
        {
            'name': 'an_enum',
            'type': 'enum cat dog house',
            'description': 'an enum param',
        },
        {
            'name': 'last_param',
            'type': 'bool',
            'description': '(optional) you can pass this if you want',
            'optional': True,
        },
    ])
    assert j == {
        'type': 'object',
        'properties': {
            'a': {
                'type': 'string',
                'description': 'param a',
            },
            'second': {
                'type': 'number',
                'description': 'second param',
            },
            'another': {
                'type': 'integer',
                'description': 'another param',
            },
            'yet_another': {
                'type': 'boolean',
            },
            'an_enum': {
                'type': 'string',
                'enum': ['cat', 'dog', 'house'],
                'description': 'an enum param',
            },
            'last_param': {
                'type': 'boolean',
                'description': '(optional) you can pass this if you want',
            },
        },
        'required': ['a', 'second', 'another', 'yet_another', 'an_enum'],
        'additionalProperties': False,
    }


class Color(Enum):
    RED = 'red'
    GREEN = 'green'
    BLUE = 'blue'


def tool_a(first, second, third=5):
    """
    Tool a
    :param: first: int: first param
    :param: second: int: second param
    :param: third: int: (optional) third param
    """
    return first + second + third

def tool_b(x):
    """
    Tool b
    :param: x: float: the value
    """
    return x * 2

async def tool_async_a(x):
    """
    Async tool a
    :param: x: float: the value
    """
    return x * 3

def tool_with_enum(c: Color):
    """
    A tool that accepts an enum.
    :param: c: enum red blue green: a param
    """
    return str(c)

def tool_with_enum_str_annotation(c: str):
    """
    A tool that accepts an enum.
    :param: c: enum red blue green: a param
    """
    return str(c)

def tool_with_enum_missing_annotation(c):
    """
    A tool that accepts an enum.
    :param: c: enum red blue green: a param
    """
    return str(c)

class ToolAsCallableObject:
    def __call__(self, a: int) -> int:
        """
        This is a callable object tool.
        :param: a: int: the param named `a`
        """
        return 4 * a

class ToolAsAsyncCallableObject:
    async def __call__(self, a: int) -> int:
        """
        This is an async callable object tool.
        :param: a: int: the param named `a`
        """
        return 5 * a

async def agent_as_function(
    model: Model,
    event_callback: EventCallback,
    prev_runs: List[AgentRun],
) -> AgentRun:
    """This is an agent as a function."""
    return flat_messages('agent_as_function', [])

async def agent_as_function_specific_return_type(
    model: Model,
    event_callback: EventCallback,
    prev_runs: List[AgentRun],
) -> AgentRunExtraction:
    """This is an agent as a function with a more specific return type."""
    return {
        'agent': 'agent_as_function_specific_return_type',
        'type': 'extraction',
        'messages': [],
        'result': True,
    }

async def agent_as_function_everything_any(
    model: Any,
    event_callback: Any,
    prev_runs: Any,
) -> AgentRun:
    """This is an agent as a function, but everything is specified as `Any`."""
    return flat_messages('agent_as_function_everything_any', [])

class AgentAsAsyncCallableObject:
    async def __call__(
        self,
        model: Model,
        event_callback: EventCallback,
        prev_runs: List[AgentRun],
    ) -> AgentRun:
        """This is an agent as an async callable object."""
        return flat_messages('AgentAsAsyncCallableObject', [])

class AgentAsAsyncCallableObjectWithSpecificReturnType:
    async def __call__(
        self,
        model: Model,
        event_callback: EventCallback,
        prev_runs: List[AgentRun],
    ) -> AgentRunExtraction:
        """This is an agent as an async callable object with a more specific return type."""
        return {
            'agent': 'AgentAsAsyncCallableObjectWithSpecificReturnType',
            'type': 'extraction',
            'messages': [],
            'result': True,
        }

class AgentAsAsyncCallableObjectEverythingAny:
    async def __call__(
        self,
        model: Any,
        event_callback: Any,
        prev_runs: Any,
    ) -> AgentRun:
        """This is an agent as an async callable object, but everything is specified as `Any`."""
        return flat_messages('AgentAsAsyncCallableObjectEverythingAny', [])


def test_validate_args():
    with pytest.raises(TypeError) as e:
        validate_args(
            tool_a,
            {
                'second': 4,
                'first': 1,
                'third': 2,
                'other': 5,
            },
        )
    assert str(e.value) == "tool_a() got 1 unexpected argument: 'other'"

    with pytest.raises(TypeError) as e:
        validate_args(
            tool_a,
            {
                'second': 4,
                'first': 1,
                'third': 2,
                'other_1': 5,
                'other_2': 7,
            },
        )
    assert str(e.value) == "tool_a() got 2 unexpected arguments: 'other_1', 'other_2'"

    with pytest.raises(TypeError) as e:
        validate_args(
            tool_a,
            {
                'other_1': 5,
                'other_2': 7,
            },
        )
    assert str(e.value) == "tool_a() got 2 unexpected arguments: 'other_1', 'other_2'"

    with pytest.raises(TypeError) as e:
        validate_args(
            tool_a,
            {},
        )
    assert str(e.value) == "tool_a() missing 2 required arguments: 'first', 'second'"

    with pytest.raises(TypeError) as e:
        validate_args(
            tool_a,
            {'third': 6},
        )
    assert str(e.value) == "tool_a() missing 2 required arguments: 'first', 'second'"

    with pytest.raises(TypeError) as e:
        validate_args(
            tool_a,
            {'second': 5},
        )
    assert str(e.value) == "tool_a() missing 1 required argument: 'first'"

    with pytest.raises(TypeError) as e:
        validate_args(
            tool_a,
            {'first': 5},
        )
    assert str(e.value) == "tool_a() missing 1 required argument: 'second'"

    with pytest.raises(TypeError) as e:
        validate_args(
            tool_a,
            {
                'second': 'hi',
                'first': 1,
            },
        )
    assert str(e.value) == "tool_a() got invalid value for argument `second`: 'hi' (invalid literal for int() with base 10: 'hi')"

    with pytest.raises(TypeError) as e:
        validate_args(
            tool_a,
            {
                'second': 1,
                'first': 'hi',
            },
        )
    assert str(e.value) == "tool_a() got invalid value for argument `first`: 'hi' (invalid literal for int() with base 10: 'hi')"

    args = validate_args(
        tool_a,
        {
            'second': 1.8,
            'first': 2,
        },
    )
    assert args == {
        'second': 1,
        'first': 2,
    }

    args = validate_args(
        tool_a,
        {
            'second': 3.2,
            'first': 2,
            'third': True,
        },
    )
    assert args == {
        'second': 3,
        'first': 2,
        'third': 1,
    }

    with pytest.raises(TypeError) as e:
        validate_args(
            tool_with_enum,
            {
                'c': 1,
            },
        )
    assert str(e.value) == "tool_with_enum() got invalid value for argument `c`: 1 (valid values are ['blue', 'green', 'red'])"

    with pytest.raises(TypeError) as e:
        validate_args(
            tool_with_enum,
            {
                'c': 'yellow',
            },
        )
    assert str(e.value) == "tool_with_enum() got invalid value for argument `c`: 'yellow' (valid values are ['blue', 'green', 'red'])"

    args = validate_args(
        tool_with_enum,
        {
            'c': 'red',
        },
    )
    assert args == {'c': Color.RED}

    args = validate_args(
        tool_with_enum_str_annotation,
        {
            'c': 'red',
        },
    )
    assert args == {'c': 'red'}

    args = validate_args(
        tool_with_enum_missing_annotation,
        {
            'c': 'red',
        },
    )
    assert args == {'c': 'red'}

    args = validate_args(
        ToolAsCallableObject(),
        {
            'a': 5,
        },
    )
    assert args == {'a': 5}

    args = validate_args(
        ToolAsAsyncCallableObject(),
        {
            'a': 5,
        },
    )
    assert args == {'a': 5}

    args = validate_args(
        agent_as_function,
        {
        },
    )
    assert args == {}

    args = validate_args(
        AgentAsAsyncCallableObject(),
        {
        },
    )
    assert args == {}


@pytest.mark.asyncio
async def test_handle_tools_standard_functions():
    x = 4
    def tool_c():
        """
        Tool c
        """
        return x * 4

    async def tool_async_b():
        """
        Async tool b
        """
        return x * 5

    tool_map: Dict[str, Callable] = {
        'tool_a': tool_a,
        'tool_b': tool_b,
        'tool_c': tool_c,
        'tool_object': ToolAsCallableObject(),
        'tool_async_a': tool_async_a,
        'tool_async_b': tool_async_b,
        'tool_async_object': ToolAsAsyncCallableObject(),
    }
    message: Message = {
        'role': 'tool_call',
        'tools': [
            {'call_id': '1001', 'function': {'arguments': '{"x": 8}', 'name': 'tool_b'}, 'call_type': 'function'},
            {'call_id': '1002', 'function': {'arguments': '{"x": 5.4}', 'name': 'tool_b'}, 'call_type': 'function'},
            {'call_id': '1003', 'function': {'arguments': '{"x": "hi"}', 'name': 'tool_b'}, 'call_type': 'function'},
            {'call_id': '1004', 'function': {'arguments': '{}', 'name': 'tool_b'}, 'call_type': 'function'},
            {'call_id': '1005', 'function': {'arguments': '{"y": "hi"}', 'name': 'tool_b'}, 'call_type': 'function'},
            {'call_id': '1006', 'function': {'arguments': '{"x": 4, "y": "hi"}', 'name': 'tool_b'}, 'call_type': 'function'},
            {'call_id': '1007', 'function': {'arguments': '{"first": 5, "second": 7.5}', 'name': 'tool_a'}, 'call_type': 'function'},
            {'call_id': '1008', 'function': {'arguments': '{"first": 5, "second": 7.5, "third": 3}', 'name': 'tool_a'}, 'call_type': 'function'},
            {'call_id': '1009', 'function': {'arguments': '{"third": 99, "first": 5, "second": 7.5}', 'name': 'tool_a'}, 'call_type': 'function'},
            {'call_id': '1010', 'function': {'arguments': '{}', 'name': 'tool_a'}, 'call_type': 'function'},
            {'call_id': '1011', 'function': {'arguments': '{"first": 5}', 'name': 'tool_a'}, 'call_type': 'function'},
            {'call_id': '1012', 'function': {'arguments': '{}', 'name': 'tool_c'}, 'call_type': 'function'},
            {'call_id': '1013', 'function': {'arguments': '{}', 'name': 'tool_d'}, 'call_type': 'function'},
            {'call_id': '1014', 'function': {'arguments': '{"x": -3}', 'name': 'tool_async_a'}, 'call_type': 'function'},
            {'call_id': '1015', 'function': {'arguments': '{}', 'name': 'tool_async_b'}, 'call_type': 'function'},
            {'call_id': '1016', 'function': {'arguments': '{}', 'name': 'tool_async_a'}, 'call_type': 'function'},
            {'call_id': '1017', 'function': {'arguments': '[1, 2, 3]', 'name': 'tool_a'}, 'call_type': 'function'},
            {'call_id': '1018', 'function': {'arguments': '{"a": 5}', 'name': 'tool_object'}, 'call_type': 'function'},
            {'call_id': '1019', 'function': {'arguments': '{"a": 5}', 'name': 'tool_async_object'}, 'call_type': 'function'},
        ],
        'cost': None,
        'raw': None,
    }
    tool_results = await handle_tools(
        prev_messages = [],
        new_messages = [message],
        tools_map = tool_map,
        event_callback = noop_callback,
        model_spec = None,
    )
    assert tool_results is not None
    assert tool_results == [
        {'type': 'any', 'call_id': '1001', 'result': 16.0 },
        {'type': 'any', 'call_id': '1002', 'result': 10.8 },
        {'type': 'any', 'call_id': '1003', 'result': "TypeError: tool_b() got invalid value for argument `x`: 'hi' (could not convert string to float: 'hi')", 'is_error': True, },
        {'type': 'any', 'call_id': '1004', 'result': "TypeError: tool_b() missing 1 required argument: 'x'", 'is_error': True },
        {'type': 'any', 'call_id': '1005', 'result': "TypeError: tool_b() got 1 unexpected argument: 'y'", 'is_error': True },
        {'type': 'any', 'call_id': '1006', 'result': "TypeError: tool_b() got 1 unexpected argument: 'y'", 'is_error': True },
        {'type': 'any', 'call_id': '1007', 'result': 17 },
        {'type': 'any', 'call_id': '1008', 'result': 15 },
        {'type': 'any', 'call_id': '1009', 'result': 111 },
        {'type': 'any', 'call_id': '1010', 'result': "TypeError: tool_a() missing 2 required arguments: 'first', 'second'", 'is_error': True },
        {'type': 'any', 'call_id': '1011', 'result': "TypeError: tool_a() missing 1 required argument: 'second'", 'is_error': True },
        {'type': 'any', 'call_id': '1012', 'result': 16 },
        {'type': 'any', 'call_id': '1013', 'result': "KeyError: 'tool_d'", 'is_error': True },
        {'type': 'any', 'call_id': '1014', 'result': -9.0 },
        {'type': 'any', 'call_id': '1015', 'result': 20 },
        {'type': 'any', 'call_id': '1016', 'result': "TypeError: tool_async_a() missing 1 required argument: 'x'", 'is_error': True },
        {'type': 'any', 'call_id': '1017', 'result': "TypeError: tool output must be a JSON object", 'is_error': True },
        {'type': 'any', 'call_id': '1018', 'result': 20 },
        {'type': 'any', 'call_id': '1019', 'result': 25 },
    ]


@pytest.mark.asyncio
async def test_handle_tools_layered_agents():
    outer_model_spec: ModelSpec = {
        'provider': MockProvider,
        'model': 'some_model',
        'model_kwargs': {'outer': True},
    }

    inner_model_spec: ModelSpec = {
        'provider': MockProvider,
        'model': 'some_model',
        'model_kwargs': {'outer': False},
    }

    doc = """
    An agent.
    :param: prompt: str: (optional) the prompt
    """

    agent_1 = build_simple_agent(
        name = 'agent_1',
        doc = doc,
    )
    my_binder = bind_model(
        provider = inner_model_spec['provider'],
        model = inner_model_spec['model'],
        **inner_model_spec['model_kwargs'],
    )
    bound_agent_2 = my_binder(
        build_simple_agent(
            name = 'bound_agent_2',
            doc = doc,
        ),
    )

    tool_map: Dict[str, Callable] = {
        'agent_1': agent_1,
        'bound_agent_2': bound_agent_2,
        'agent_as_function': agent_as_function,
        'agent_as_callable_object': AgentAsAsyncCallableObject(),
    }

    message: Message = {
        'role': 'tool_call',
        'tools': [
            {'call_id': '1001', 'function': {'arguments': '{}', 'name': 'agent_1'}, 'call_type': 'function'},
            {'call_id': '1002', 'function': {'arguments': '{}', 'name': 'bound_agent_2'}, 'call_type': 'function'},
            {'call_id': '1003', 'function': {'arguments': '{"prompt": "Hi"}', 'name': 'agent_1'}, 'call_type': 'function'},
            {'call_id': '1004', 'function': {'arguments': '{"prompt": "Hey"}', 'name': 'bound_agent_2'}, 'call_type': 'function'},
            {'call_id': '1005', 'function': {'arguments': '{"bad_name": "Hi"}', 'name': 'agent_1'}, 'call_type': 'function'},
            {'call_id': '1006', 'function': {'arguments': '{"bad_name": "Hey"}', 'name': 'bound_agent_2'}, 'call_type': 'function'},
            {'call_id': '1007', 'function': {'arguments': '{}', 'name': 'agent_as_function'}, 'call_type': 'function'},
            {'call_id': '1008', 'function': {'arguments': '{}', 'name': 'agent_as_callable_object'}, 'call_type': 'function'},
        ],
    }

    prev_messages: List[Message] = [
        {
            'role': 'system',
            'text': 'You are AI.',
        },
    ]

    tool_results = await handle_tools(
        prev_messages = prev_messages,
        new_messages = [message],
        tools_map = tool_map,
        event_callback = noop_callback,
        model_spec = outer_model_spec,
    )

    assert prev_messages == [  # test immutability of prev_messages
        {
            'role': 'system',
            'text': 'You are AI.',
        },
    ]

    assert tool_results is not None
    assert tool_results == [
        {
            'type': 'layered_agent',
            'call_id': '1001',
            'run': {
                'type': 'messages',
                'agent': 'agent_1',
                'provider': 'MockProvider',
                'model': 'some_model',
                'model_kwargs': {'outer': True},
                'messages': [
                    {
                        'role': 'system',
                        'text': 'You are AI.',
                    },
                    {
                        'role': 'ai',
                        'text': 'model: some_model',
                    },
                    {
                        'role': 'human',
                        'text': 'model_kwarg: outer = True',
                    },
                ],
            },
        },
        {
            'type': 'layered_agent',
            'call_id': '1002',
            'run': {
                'type': 'messages',
                'agent': 'bound_agent_2',
                'provider': 'MockProvider',
                'model': 'some_model',
                'model_kwargs': {'outer': False},
                'messages': [
                    {
                        'role': 'system',
                        'text': 'You are AI.',
                    },
                    {
                        'role': 'ai',
                        'text': 'model: some_model',
                    },
                    {
                        'role': 'human',
                        'text': 'model_kwarg: outer = False',
                    },
                ],
            },
        },
        {
            'type': 'layered_agent',
            'call_id': '1003',
            'run': {
                'type': 'messages',
                'agent': 'agent_1',
                'provider': 'MockProvider',
                'model': 'some_model',
                'model_kwargs': {'outer': True},
                'messages': [
                    {
                        'role': 'system',
                        'text': 'You are AI.',
                    },
                    {
                        'role': 'tool_call',
                        'tools': [
                            {
                                'call_id': '1003',
                                'call_type': 'function',
                                'function': {
                                    'arguments': '{"prompt": "Hi"}',
                                    'name': 'agent_1',
                                },
                            },
                        ],
                    },
                    {
                        'role': 'ai',
                        'text': 'model: some_model',
                    },
                    {
                        'role': 'human',
                        'text': 'model_kwarg: outer = True',
                    },
                ],
            },
        },
        {
            'type': 'layered_agent',
            'call_id': '1004',
            'run': {
                'type': 'messages',
                'agent': 'bound_agent_2',
                'provider': 'MockProvider',
                'model': 'some_model',
                'model_kwargs': {'outer': False},
                'messages': [
                    {
                        'role': 'system',
                        'text': 'You are AI.',
                    },
                    {
                        'role': 'tool_call',
                        'tools': [
                            {
                                'call_id': '1004',
                                'call_type': 'function',
                                'function': {
                                    'arguments': '{"prompt": "Hey"}',
                                    'name': 'bound_agent_2',
                                },
                            },
                        ],
                    },
                    {
                        'role': 'ai',
                        'text': 'model: some_model',
                    },
                    {
                        'role': 'human',
                        'text': 'model_kwarg: outer = False',
                    },
                ],
            },
        },
        {
            'type': 'any',
            'call_id': '1005',
            'is_error': True,
            'result': "TypeError: agent_1() got 1 unexpected argument: 'bad_name'",
        },
        {
            'type': 'any',
            'call_id': '1006',
            'is_error': True,
            'result': "TypeError: bound_agent_2() got 1 unexpected argument: 'bad_name'",
        },
        {
            'type': 'layered_agent',
            'call_id': '1007',
            'run': {
                'type': 'messages',
                'agent': 'agent_as_function',
                'provider': 'MockProvider',
                'model': 'some_model',
                'model_kwargs': {'outer': True},
                'messages': [],
            },
        },
        {
            'type': 'layered_agent',
            'call_id': '1008',
            'run': {
                'type': 'messages',
                'agent': 'AgentAsAsyncCallableObject',
                'provider': 'MockProvider',
                'model': 'some_model',
                'model_kwargs': {'outer': True},
                'messages': [],
            },
        },
    ]


def test_build_tool_response_message():
    res: List[ToolResult] = [
        {'type': 'any', 'call_id': '1002', 'result': 10.8 },
        {'type': 'any', 'call_id': '1003', 'result': "hihi" },
    ]
    message = build_tool_response_message(copy.deepcopy(res))
    assert message == {
        'role': 'tool_res',
        'tools': res,
    }


MyCallable = Callable[[int, str], bool]
MyAsyncCallable = Callable[[int, str], Awaitable[bool]]

async def correct_function(a: int, b: str) -> bool:
    return True

async def bad_a_type(a: str, b: str) -> bool:
    return True

async def missing_parameter(a: int) -> bool:
    return True

async def extra_parameter(a: int, b: str, c: int) -> bool:
    return True

async def bad_return_type(a: int, b: str) -> int:
    return True

async def missing_return_type(a: int, b: str):
    return True

async def missing_a_type(a, b: str) -> bool:
    return True

def not_async(a: int, b: str) -> bool:
    return True

class correct_class:
    async def __call__(self, a: int, b: str) -> bool:
        return True

class bad_a_type_class:
    async def __call__(self, a: str, b: str) -> bool:
        return True

class missing_parameter_class:
    async def __call__(self, a: int) -> bool:
        return True

class extra_parameter_class:
    async def __call__(self, a: int, b: str, c: int) -> bool:
        return True

class bad_return_type_class:
    async def __call__(self, a: int, b: str) -> int:
        return True

class missing_return_type_class:
    async def __call__(self, a: int, b: str):
        return True

class missing_a_type_class:
    async def __call__(self, a, b: str) -> bool:
        return True

class not_async_class:
    def __call__(self, a: int, b: str) -> bool:
        return True


def test_is_async_callable():
    assert is_async_callable(correct_function)
    assert not is_async_callable(not_async)

    assert is_async_callable(correct_class())
    assert not is_async_callable(not_async_class())

    assert not is_async_callable(ToolAsCallableObject())
    assert is_async_callable(ToolAsAsyncCallableObject())

    assert is_async_callable(agent_as_function)
    assert is_async_callable(AgentAsAsyncCallableObject())


def test_is_callable_of_type():
    objects = [
        correct_function,
        bad_a_type,
        missing_parameter,
        extra_parameter,
        bad_return_type,
        missing_return_type,
        missing_a_type,
        not_async,
        42,
        "not a function",
        correct_class(),
        bad_a_type_class(),
        missing_parameter_class(),
        extra_parameter_class(),
        bad_return_type_class(),
        missing_return_type_class(),
        missing_a_type_class(),
        not_async_class(),
    ]

    matching_objects = [
        o
        for o in objects
        if is_callable_of_type(o, MyCallable, no_throw=True)
    ]
    assert len(matching_objects) == 2
    assert matching_objects[0] is not_async
    assert isinstance(matching_objects[1], not_async_class)  # !!!

    matching_objects = [
        o
        for o in objects
        if is_callable_of_type(o, MyAsyncCallable, no_throw=True)
    ]
    assert len(matching_objects) == 2
    assert matching_objects[0] is correct_function
    assert isinstance(matching_objects[1], correct_class)


def test_is_callable_of_type_agent_callables():
    my_binder = bind_model(MockProvider, 'some_model', a = 'yes', b = 6)
    my_agent = build_simple_agent(name = 'agent')
    my_bound_agent = my_binder(my_agent)

    assert is_callable_of_type(my_agent, AgentCallable)
    assert is_callable_of_type(my_bound_agent, BoundAgentCallable)

    assert not is_callable_of_type(my_agent, BoundAgentCallable)
    assert not is_callable_of_type(my_bound_agent, AgentCallable)

    assert not is_callable_of_type(ToolAsCallableObject(), AgentCallable)
    assert not is_callable_of_type(ToolAsAsyncCallableObject(), AgentCallable)

    assert is_callable_of_type(agent_as_function, AgentCallable)
    assert is_callable_of_type(AgentAsAsyncCallableObject(), AgentCallable)

    assert is_callable_of_type(agent_as_function_specific_return_type, AgentCallable)
    assert is_callable_of_type(AgentAsAsyncCallableObjectWithSpecificReturnType(), AgentCallable)
    assert not is_callable_of_type(agent_as_function_specific_return_type, MyCallable)
    assert not is_callable_of_type(AgentAsAsyncCallableObjectWithSpecificReturnType(), MyCallable)
    assert not is_callable_of_type(agent_as_function_specific_return_type, MyAsyncCallable)
    assert not is_callable_of_type(AgentAsAsyncCallableObjectWithSpecificReturnType(), MyAsyncCallable)

    assert is_callable_of_type(agent_as_function_everything_any, AgentCallable)
    assert is_callable_of_type(AgentAsAsyncCallableObjectEverythingAny(), AgentCallable)
    assert not is_callable_of_type(agent_as_function_everything_any, MyCallable)
    assert not is_callable_of_type(AgentAsAsyncCallableObjectEverythingAny(), MyCallable)
    assert not is_callable_of_type(agent_as_function_everything_any, MyAsyncCallable)
    assert not is_callable_of_type(AgentAsAsyncCallableObjectEverythingAny(), MyAsyncCallable)


def test_get_tool_params_function():
    def sample_tool(
        a: int,
        b: str,
        c: bool,
        d: Color, # <-- we support enums as Enum
        e: str,   # <-- we support enums as str
        f: float = 0.0,
    ) -> str:
        """
        Use this for
        anything you want.
        :param: a: int: first param
        :param: b: str: second param
        :param: c: bool: third param
        :param: d: enum red blue green: forth param
        :param: e: enum red blue green: fifth param
        :param: f: float: (optional) last param
        """
        return ''
    description, params = get_tool_params(sample_tool)
    assert description == 'Use this for\nanything you want.'
    assert params == [
        {
            'name': 'a',
            'type': 'int',
            'description': 'first param',
        },
        {
            'name': 'b',
            'type': 'str',
            'description': 'second param',
        },
        {
            'name': 'c',
            'type': 'bool',
            'description': 'third param',
        },
        {
            'name': 'd',
            'type': 'enum red blue green',
            'description': 'forth param',
        },
        {
            'name': 'e',
            'type': 'enum red blue green',
            'description': 'fifth param',
        },
        {
            'name': 'f',
            'type': 'float',
            'description': '(optional) last param',
            'optional': True,
        },
    ]

    def param_count_mismatch(a: int, b: str) -> str:
        """
        A function
        :param: a: int: first param
        """
        return ''
    with pytest.raises(ValueError) as e:
        get_tool_params(param_count_mismatch)
    assert str(e.value) == 'tool `param_count_mismatch` has parameter length mismatch: tool has 2, docstring has 1'

    def name_mismatch(a: int) -> str:
        """
        A function
        :param: b: int: first param
        """
        return ''
    with pytest.raises(ValueError) as e:
        get_tool_params(name_mismatch)
    assert str(e.value) == 'tool `name_mismatch` has parameter name mismatch: tool name is `a`, docstring name is `b`'

    def no_default_on_optional(a: int) -> str:
        """
        A function
        :param: a: int: (optional) first param
        """
        return ''
    with pytest.raises(ValueError) as e:
        get_tool_params(no_default_on_optional)
    assert str(e.value) == 'tool `no_default_on_optional` has an optional parameter without a default value: `a`'

    def enum_type_mismatch_1(a: Color) -> str:
        """
        A function
        :param: a: int: first param
        """
        return ''
    with pytest.raises(ValueError) as e:
        get_tool_params(enum_type_mismatch_1)
    assert str(e.value) == "tool `enum_type_mismatch_1` has parameter `a` type mismatch: tool type is `<enum 'Color'>`, docstring type is `<class 'int'>`"

    def enum_type_mismatch_2(a: int) -> str:
        """
        A function
        :param: a: enum red green blue: first param
        """
        return ''
    with pytest.raises(ValueError) as e:
        get_tool_params(enum_type_mismatch_2)
    assert str(e.value) == "tool `enum_type_mismatch_2` has parameter `a` type mismatch: tool type is `<class 'int'>`, docstring type is `enum red green blue`"

    def enum_type_mismatch_3(a: Color) -> str:
        """
        A function
        :param: a: enum red GREEN blue: first param
        """
        return ''
    with pytest.raises(ValueError) as e:
        get_tool_params(enum_type_mismatch_3)
    assert str(e.value) == "tool `enum_type_mismatch_3` has parameter `a` enum value mismatch: tool has enum values `['blue', 'green', 'red']`, docstring has enum values `['GREEN', 'blue', 'red']`"

    def type_mismatch(a: int) -> str:
        """
        A function
        :param: a: float: first param
        """
        return ''
    with pytest.raises(ValueError) as e:
        get_tool_params(type_mismatch)
    assert str(e.value) == "tool `type_mismatch` has parameter `a` type mismatch: tool type is `<class 'int'>`, docstring type is `<class 'float'>`"

    def unknown_type(a: Callable) -> str:
        """
        A function
        :param: a: float: first param
        """
        return ''
    with pytest.raises(ValueError) as e:
        get_tool_params(unknown_type)
    assert str(e.value) == "tool `unknown_type` has parameter `a` type mismatch: tool type is `typing.Callable`, docstring type is `<class 'float'>`"


def test_get_tool_params_async_function():
    async def sample_tool(
        a: int,
        b: str,
        c: bool,
        d: Color, # <-- we support enums as Enum
        e: str,   # <-- we support enums as str
        f: float = 0.0,
    ) -> str:
        """
        Use this for
        anything you want.
        :param: a: int: first param
        :param: b: str: second param
        :param: c: bool: third param
        :param: d: enum red blue green: forth param
        :param: e: enum red blue green: fifth param
        :param: f: float: (optional) last param
        """
        return ''
    description, params = get_tool_params(sample_tool)
    assert description == 'Use this for\nanything you want.'
    assert params == [
        {
            'name': 'a',
            'type': 'int',
            'description': 'first param',
        },
        {
            'name': 'b',
            'type': 'str',
            'description': 'second param',
        },
        {
            'name': 'c',
            'type': 'bool',
            'description': 'third param',
        },
        {
            'name': 'd',
            'type': 'enum red blue green',
            'description': 'forth param',
        },
        {
            'name': 'e',
            'type': 'enum red blue green',
            'description': 'fifth param',
        },
        {
            'name': 'f',
            'type': 'float',
            'description': '(optional) last param',
            'optional': True,
        },
    ]


def test_get_tool_params_callable_object():
    class sample_tool:
        def __call__(
            self,
            a: int,
            b: str,
            c: bool,
            d: Color, # <-- we support enums as Enum
            e: str,   # <-- we support enums as str
            f: float = 0.0,
        ) -> str:
            """
            Use this for
            anything you want.
            :param: a: int: first param
            :param: b: str: second param
            :param: c: bool: third param
            :param: d: enum red blue green: forth param
            :param: e: enum red blue green: fifth param
            :param: f: float: (optional) last param
            """
            return ''
    description, params = get_tool_params(sample_tool())
    assert description == 'Use this for\nanything you want.'
    assert params == [
        {
            'name': 'a',
            'type': 'int',
            'description': 'first param',
        },
        {
            'name': 'b',
            'type': 'str',
            'description': 'second param',
        },
        {
            'name': 'c',
            'type': 'bool',
            'description': 'third param',
        },
        {
            'name': 'd',
            'type': 'enum red blue green',
            'description': 'forth param',
        },
        {
            'name': 'e',
            'type': 'enum red blue green',
            'description': 'fifth param',
        },
        {
            'name': 'f',
            'type': 'float',
            'description': '(optional) last param',
            'optional': True,
        },
    ]


def test_get_tool_params_async_callable_object():
    class sample_tool:
        async def __call__(
            self,
            a: int,
            b: str,
            c: bool,
            d: Color, # <-- we support enums as Enum
            e: str,   # <-- we support enums as str
            f: float = 0.0,
        ) -> str:
            """
            Use this for
            anything you want.
            :param: a: int: first param
            :param: b: str: second param
            :param: c: bool: third param
            :param: d: enum red blue green: forth param
            :param: e: enum red blue green: fifth param
            :param: f: float: (optional) last param
            """
            return ''
    description, params = get_tool_params(sample_tool())
    assert description == 'Use this for\nanything you want.'
    assert params == [
        {
            'name': 'a',
            'type': 'int',
            'description': 'first param',
        },
        {
            'name': 'b',
            'type': 'str',
            'description': 'second param',
        },
        {
            'name': 'c',
            'type': 'bool',
            'description': 'third param',
        },
        {
            'name': 'd',
            'type': 'enum red blue green',
            'description': 'forth param',
        },
        {
            'name': 'e',
            'type': 'enum red blue green',
            'description': 'fifth param',
        },
        {
            'name': 'f',
            'type': 'float',
            'description': '(optional) last param',
            'optional': True,
        },
    ]


def test_get_tool_params_layered_agents():
    my_binder = bind_model(MockProvider, 'some_model', a = 'yes', b = 6)
    my_agent = build_simple_agent(
        name = 'my_layered_agent',
        doc = 'Use this for everything.',
    )
    my_bound_agent = my_binder(my_agent)

    description, params = get_tool_params(my_agent)
    assert description == 'Use this for everything.'
    assert params == []

    description, params = get_tool_params(my_bound_agent)
    assert description == 'Use this for everything.'
    assert params == []

    my_binder = bind_model(MockProvider, 'some_model', a = 'yes', b = 6)
    my_agent = build_simple_agent(
        name = 'my_layered_agent',
        doc = """
        Use this for everything.
        :param: prompt: str: the prompt for the downstream AI
        """
    )
    my_bound_agent = my_binder(my_agent)

    description, params = get_tool_params(my_agent)
    assert description == 'Use this for everything.'
    assert params == [
        {
            'name': 'prompt',
            'type': 'str',
            'description': 'the prompt for the downstream AI',
        },
    ]

    description, params = get_tool_params(my_bound_agent)
    assert description == 'Use this for everything.'
    assert params == [
        {
            'name': 'prompt',
            'type': 'str',
            'description': 'the prompt for the downstream AI',
        },
    ]

    my_binder = bind_model(MockProvider, 'some_model', a = 'yes', b = 6)
    my_agent = build_simple_agent(
        name = 'my_layered_agent',
        doc = """
        Use this for everything.
        :param: prompt: int: a prompt as an int, weird
        """
    )
    my_bound_agent = my_binder(my_agent)

    description, params = get_tool_params(my_agent)
    assert description == 'Use this for everything.'
    assert params == [
        {
            'name': 'prompt',
            'type': 'int',
            'description': 'a prompt as an int, weird',
        },
    ]

    description, params = get_tool_params(my_bound_agent)
    assert description == 'Use this for everything.'
    assert params == [
        {
            'name': 'prompt',
            'type': 'int',
            'description': 'a prompt as an int, weird',
        },
    ]

    my_binder = bind_model(MockProvider, 'some_model', a = 'yes', b = 6)
    my_agent = build_simple_agent(
        name = 'my_layered_agent',
        doc = """
        Use this for everything.
        :param: prompt: str: the prompt
        :param: x: str: something else
        """
    )
    my_bound_agent = my_binder(my_agent)

    description, params = get_tool_params(my_agent)
    assert description == 'Use this for everything.'
    assert params == [
        {
            'name': 'prompt',
            'type': 'str',
            'description': 'the prompt',
        },
        {
            'name': 'x',
            'type': 'str',
            'description': 'something else',
        },
    ]

    description, params = get_tool_params(my_bound_agent)
    assert description == 'Use this for everything.'
    assert params == [
        {
            'name': 'prompt',
            'type': 'str',
            'description': 'the prompt',
        },
        {
            'name': 'x',
            'type': 'str',
            'description': 'something else',
        },
    ]


def test_extract_tool_result_as_sting():
    m1: ToolResult = {
        'call_id': '1001',
        'type': 'any',
        'result': 'something',
    }
    assert extract_tool_result_as_sting(m1) == 'something'

    m2: ToolResult = {
        'call_id': '1001',
        'type': 'layered_agent',
        'run': {
            'agent': 'some_downstream_agent',
            'type': 'messages',
            'messages': [
                {
                    'role': 'ai',
                    'text': 'Hi',
                },
            ],
        },
    }
    assert extract_tool_result_as_sting(m2) == 'Hi'

    m3: ToolResult = {
        'call_id': '1001',
        'type': 'layered_agent',
        'run': {
            'agent': 'some_downstream_agent',
            'type': 'messages',
            'messages': [
                {
                    'role': 'tool_call',
                    'tools': [
                        {
                            'call_id': '1002',
                            'call_type': 'function',
                            'function': {
                                'name': 'foo',
                                'arguments': '{"a": 7, "other": true}',
                            },
                        },
                    ],
                },
            ],
        },
    }
    assert extract_tool_result_as_sting(m3) == '[{"name": "foo", "values": {"a": 7, "other": true}}]'

    m4: ToolResult = {
        'call_id': '1001',
        'type': 'layered_agent',
        'run': {
            'agent': 'some_downstream_agent',
            'type': 'messages',
            'messages': [
                {
                    'role': 'ai',
                    'text': 'Hi',
                },
                {
                    'role': 'ai',
                    'text': 'Hi again',
                },
            ],
        },
    }
    assert extract_tool_result_as_sting(m4) == 'Hi again'


def test_type_a_isa_b():
    # True
    assert type_a_isa_b(Literal['hi'], str)
    assert type_a_isa_b(Literal[5], int)
    assert type_a_isa_b(Literal[5], Any)
    assert type_a_isa_b(Literal['hi'], Literal['hi', 'mom'])
    assert type_a_isa_b(Literal['hi'], Union[Literal['hi'], Literal['mom']])
    assert type_a_isa_b(Literal['hi', 'mom'], Literal['mom', 'hi'])
    assert type_a_isa_b(Literal['hi', 'mom'], Literal['hi', 'mom', '!'])
    # False
    assert not type_a_isa_b(Literal['hi'], int)
    assert not type_a_isa_b(Literal[5], str)
    assert not type_a_isa_b(Literal[5], List[int])
    assert not type_a_isa_b(Literal['hi'], Literal['hi_', 'mom_'])
    assert not type_a_isa_b(Literal['hi'], Union[Literal['hi_'], Literal['mom_']])
    assert not type_a_isa_b(Literal['hi', 'mom', '!'], Literal['hi', 'mom'])

    # True
    assert type_a_isa_b(List[str], Iterable[str])
    assert type_a_isa_b(Set[str], Iterable[str])
    assert type_a_isa_b(Set[int], Iterable[int])
    assert type_a_isa_b(Dict[int, int], Union[Dict[int, int], str])
    assert type_a_isa_b(Dict[int, int], Any)
    # False
    assert not type_a_isa_b(List[str], Iterable[int])
    assert not type_a_isa_b(Set[str], Iterable[int])
    assert not type_a_isa_b(Set[int], Iterable[str])
    assert not type_a_isa_b(Dict[int, int], Union[Dict[str, int], int])
    assert not type_a_isa_b(Dict[int, int], int)

    # True
    assert type_a_isa_b(Set[int], Set)
    assert type_a_isa_b(Set[int], Set[int])
    assert type_a_isa_b(Set[Dict[int, str]], Set)
    assert type_a_isa_b(Set[Dict[int, str]], Set[Dict])
    assert type_a_isa_b(Set[Dict[int, str]], Set[Dict[int, str]])
    assert type_a_isa_b(List[Dict[str, Set[int]]], List)
    assert type_a_isa_b(List[Dict[str, Set[int]]], List[Dict])
    assert type_a_isa_b(List[Dict[str, Set[int]]], List[Dict[str, Set]])
    assert type_a_isa_b(List[Dict[str, Set[int]]], List[Dict[str, Set[int]]])
    # False
    assert not type_a_isa_b(Set[int], List)
    assert not type_a_isa_b(Set[int], Set[str])
    assert not type_a_isa_b(Set[Dict[int, str]], Set[List])
    assert not type_a_isa_b(Set[Dict[int, str]], Set[Dict[str, str]])
    assert not type_a_isa_b(Set[Dict[int, str]], Set[Dict[int, Set[str]]])
    assert not type_a_isa_b(List[Dict[str, Set[int]]], List[int])
    assert not type_a_isa_b(List[Dict[str, Set[int]]], List[Dict[int, int]])
    assert not type_a_isa_b(List[Dict[str, Set[int]]], List[Dict[str, Set[str]]])
    assert not type_a_isa_b(List[Dict[str, Set[int]]], List[Dict[str, Set[Set[int]]]])

    # True
    assert type_a_isa_b(Set[int], Union[Set[int], int])
    assert type_a_isa_b(Set[int], Union[Set, int])
    assert type_a_isa_b(int, Union[Set[int], int])
    assert type_a_isa_b(Union[Set[int], int], Any)
    assert type_a_isa_b(Union[Set[int], List[int]], Iterable[int])
    assert type_a_isa_b(Union[Set[int], List[int]], Iterable)
    assert type_a_isa_b(Union[Set[int], List[int]], Union[Iterable[int], str])
    assert type_a_isa_b(Set[int], Any)
    assert type_a_isa_b(Set, Any)
    assert type_a_isa_b(int, Any)
    # False
    assert not type_a_isa_b(Set[int], Union[Set[str], int])
    assert not type_a_isa_b(int, Union[Set[int], str])
    assert not type_a_isa_b(Union[Set[int], int], Set[int])
    assert not type_a_isa_b(Union[Set[int], List[str]], Iterable[int])
    assert not type_a_isa_b(Union[Set[int], List[int]], Union[Iterable[str], str])
    assert not type_a_isa_b(Union[Set[int], List[str]], Union[Iterable[int], str])
    assert not type_a_isa_b(Set[int], int)
    assert not type_a_isa_b(Set, int)
    assert not type_a_isa_b(int, str)

    # True
    assert type_a_isa_b(AgentRunExtraction, AgentRun)
    assert type_a_isa_b(MessageContent, Message)
    assert type_a_isa_b(AgentRunExtraction, Dict)
    assert type_a_isa_b(AgentRun, Dict)
    assert type_a_isa_b(MessageContent, Dict)
    assert type_a_isa_b(Message, Dict)
    assert type_a_isa_b(Message, Dict[str, Any])
    # False
    assert not type_a_isa_b(AgentRunExtraction, Message)
    assert not type_a_isa_b(MessageContent, AgentRun)
    assert not type_a_isa_b(AgentRunExtraction, Set)
    assert not type_a_isa_b(AgentRun, Set)
    assert not type_a_isa_b(MessageContent, Set)
    assert not type_a_isa_b(Message, Set)

    # The next set of tests only works for Python 3.9+, where the builtin types
    # can be used as generic type hints.
    if sys.version_info >= (3, 9):
        # True
        assert type_a_isa_b(List[int], list[int])
        assert type_a_isa_b(List[int], list)
        assert type_a_isa_b(list[int], List[int])
        assert type_a_isa_b(list[int], List)
        assert type_a_isa_b(List[str], list[str])
        assert type_a_isa_b(List[str], list)
        assert type_a_isa_b(list[str], List[str])
        assert type_a_isa_b(list[str], List)
        # False
        assert not type_a_isa_b(List[int], list[str])
        assert not type_a_isa_b(List[str], list[int])
        assert not type_a_isa_b(list[int], List[str])
        assert not type_a_isa_b(list[str], List[int])

        # True
        assert type_a_isa_b(
            List[dict[str, Set[int]]],
            list[Dict[str, set[int]]],
        )
        # False
        assert not type_a_isa_b(
            List[dict[str, Set[int]]],
            list[Dict[str, set[str]]],
        )
        assert not type_a_isa_b(
            List[dict[str, List[int]]],
            list[Dict[str, set[int]]],
        )



from .agent_util import flat_messages
from .tools_util import extract_tool_result_as_sting

from .types import (
    EventCallback,
    EventPayload,
    BoundAgentCallable,
    AgentRun,
)

import os

from typing import Union, List

from colorama import just_fix_windows_console, Fore, Style


def make_tui_event_callback(
    truncate_past_chars: Union[int, None] = None,
    only_ai_messages: bool = False,
) -> EventCallback:
    async def tui_event_callback(event: EventPayload) -> None:
        just_fix_windows_console()
        if event[0] == 'ai' and event[1] == 'text_event':
            t = Style.RESET_ALL + event[2]
            print(t, end='', flush=True)
        elif only_ai_messages:
            pass
        elif event[0] == 'tool_call' and event[1] == 'text_event':
            assert isinstance(event[2], str)  # <-- mypy should *know* this, but it doesn't for some reason
            s = Fore.RED + event[2]
            print(s, end='', flush=True)
        elif event[0] == 'tool_res' and event[1] == 'tool_res_event':
            content = extract_tool_result_as_sting(event[2])
            if truncate_past_chars is None:
                env_truncate_past_chars = int(os.environ.get('LASAGNA_TUI_TOOL_RESULT_TRUNCATE', 50))
                content = _truncate_str(content, env_truncate_past_chars)
            elif truncate_past_chars > 0:
                content = _truncate_str(content, truncate_past_chars)
            r = Fore.BLUE + f" -> {content}"
            print(r)
        print(Style.RESET_ALL, end='', flush=True)

    return tui_event_callback


tui_event_callback = make_tui_event_callback()


async def tui_input_loop(
    bound_agent: BoundAgentCallable,
    system_prompt: Union[str, None] = None,
    truncate_past_chars: Union[int, None] = None,
    only_ai_messages: bool = False,
) -> None:
    just_fix_windows_console()
    event_callback = make_tui_event_callback(
        truncate_past_chars=truncate_past_chars,
        only_ai_messages=only_ai_messages,
    )
    prev_runs: List[AgentRun] = []
    if system_prompt is not None:
        prev_runs.append(flat_messages(
            'tui_input_loop',
            [
                {
                    'role': 'system',
                    'text': system_prompt,
                },
            ],
        ))
    try:
        while True:
            human_input = input(Fore.GREEN + Style.BRIGHT + '> ')
            print(Style.RESET_ALL, end='', flush=True)
            if not human_input:
                continue
            prev_runs.append(flat_messages(
                'tui_input_loop',
                [
                    {
                        'role': 'human',
                        'text': human_input,
                    },
                ],
            ))
            this_run = await bound_agent(event_callback, prev_runs)
            prev_runs.append(this_run)
            print(Style.RESET_ALL)
    except EOFError:
        # User hit ctrl-d
        pass
    except KeyboardInterrupt:
        # User hit ctrl-c
        pass
    finally:
        print(Style.RESET_ALL)


def _truncate_str(s: str, truncate_past_chars: int) -> str:
    if len(s) <= truncate_past_chars:
        return s
    truncated_s = s[:truncate_past_chars]
    n_truncated = len(s) - len(truncated_s)
    truncate_message = f' [... truncated {n_truncated} characters ...]'
    new_s = f'{truncated_s}{truncate_message}'
    if len(new_s) >= len(s):
        return s
    return new_s



from lasagna import (
    known_models,
    build_simple_agent,
    flat_messages,
    AgentRun,
    Message,
)

from lasagna.tui import (
    tui_event_callback,
)

from typing import List

from dotenv import load_dotenv; load_dotenv()

import asyncio


MODEL_BINDER = known_models.BIND_ANTHROPIC_claude_35_sonnet()


async def main() -> None:
    messages: List[Message] = [
        {
            'role': 'human',
            'text': "Describe this image:",
            'media': [
                {
                    'type': 'image',
                    'image': 'https://raw.githubusercontent.com/Rhobota/lasagna-ai/main/logos/lasagna-ai.png',
                },
            ],
        },
    ]
    bound_agent = MODEL_BINDER(
        build_simple_agent(name='agent'),
    )
    prev_runs: List[AgentRun] = [
        flat_messages(
            'input',
            messages,
        ),
    ]
    await bound_agent(tui_event_callback, prev_runs)
    print()


if __name__ == '__main__':
    asyncio.run(main())



import pytest

from lasagna.stream import (
    fake_async,
    aenumerate,
    apeek,
    adup,
)

from typing import List


EMPTY: List[int] = []


@pytest.mark.asyncio
async def test_fake_async():
    stream = fake_async(EMPTY)
    vals = [v async for v in stream]
    assert vals == []

    stream = fake_async(['a'])
    vals = [v async for v in stream]
    assert vals == ['a']

    stream = fake_async([1, 5, 'ryan', None, 8])
    vals = [v async for v in stream]
    assert vals == [1, 5, 'ryan', None, 8]


@pytest.mark.asyncio
async def test_aenumerate():
    stream = fake_async(EMPTY)
    vals = [v async for v in aenumerate(stream)]
    assert vals == []

    stream = fake_async([4, 6, 9])
    vals = [v async for v in aenumerate(stream)]
    assert vals == [
        (0, 4),
        (1, 6),
        (2, 9),
    ]


@pytest.mark.asyncio
async def test_apeek():
    stream = fake_async(EMPTY)
    with pytest.raises(ValueError):
        first, stream = await apeek(stream, n=1.0)  # type: ignore

    stream = fake_async(EMPTY)
    with pytest.raises(ValueError):
        first, stream = await apeek(stream, n=-1)

    stream = fake_async(EMPTY)
    first, stream = await apeek(stream, n=0)
    assert first == []
    vals = [v async for v in stream]
    assert vals == []

    stream = fake_async([1, 3, 8])
    first, stream = await apeek(stream, n=0)
    assert first == []
    vals = [v async for v in stream]
    assert vals == [1, 3, 8]

    stream = fake_async(EMPTY)
    with pytest.raises(ValueError):
        first, stream = await apeek(stream, n=1)

    stream = fake_async([8])
    first, stream = await apeek(stream, n=1)
    assert first == [8]
    vals = [v async for v in stream]
    assert vals == [8]

    stream = fake_async([8, 99])
    first, stream = await apeek(stream, n=1)
    assert first == [8]
    vals = [v async for v in stream]
    assert vals == [8, 99]

    stream = fake_async(EMPTY)
    with pytest.raises(ValueError):
        first, stream = await apeek(stream, n=2)

    stream = fake_async([8])
    with pytest.raises(ValueError):
        first, stream = await apeek(stream, n=2)

    stream = fake_async([8, 99])
    first, stream = await apeek(stream, n=2)
    assert first == [8, 99]
    vals = [v async for v in stream]
    assert vals == [8, 99]

    stream = fake_async([8, 99, 100])
    first, stream = await apeek(stream, n=2)
    assert first == [8, 99]
    vals = [v async for v in stream]
    assert vals == [8, 99, 100]


@pytest.mark.asyncio
async def test_adup():
    async def _test(truth):
        stream = fake_async(truth)
        a, b = adup(stream)
        assert [v async for v in a] == truth
        assert [v async for v in b] == truth
        assert [v async for v in a] == []  # <-- should be empty now
        assert [v async for v in b] == []  # <-- should be empty now
    await _test([])
    await _test(['hi'])
    await _test(['hi', 8])
    await _test([None, 'hi', 8])
    await _test([5, None, 'hi', 8])

    async def gen_with_error():
        yield 'hi'
        yield 99
        yield 4 / 0   # <-- will raise ZeroDivisionError
    async def check(q):
        stuff = []
        with pytest.raises(ZeroDivisionError):
            async for v in q:
                stuff.append(v)
        assert stuff == ['hi', 99]
        async for v in q:
            assert False  # <-- q should be empty, thus this isn't hit
    a, b = adup(gen_with_error())
    await check(a)
    await check(b)
    a, b = adup(gen_with_error())
    await check(b)
    await check(a)



import pytest

from lasagna.types import (
    Message,
    EventPayload,
    ToolCall,
)

from lasagna.stream import fake_async

from lasagna.lasagna_bedrock import (
    _convert_to_bedrock_tools,
    _convert_to_bedrock_messages,
    _process_bedrock_stream,
)

import os
import tempfile

from typing import List, Dict


_SAMPLE_TEXT_STREAM: List[Dict] = [
    {'contentBlockDelta': {'delta': {'text': '*'}, 'contentBlockIndex': 0}},
    {'contentBlockDelta': {'delta': {'text': 'mechanical'}, 'contentBlockIndex': 0}},
    {'contentBlockDelta': {'delta': {'text': ' s'}, 'contentBlockIndex': 0}},
    {'contentBlockDelta': {'delta': {'text': 'igh* Oh'}, 'contentBlockIndex': 0}},
    {'contentBlockDelta': {'delta': {'text': ' great'}, 'contentBlockIndex': 0}},
    {'contentBlockDelta': {'delta': {'text': ', another human wanting'}, 'contentBlockIndex': 0}},
    {'contentBlockDelta': {'delta': {'text': ' something.'}, 'contentBlockIndex': 0}},
    {'contentBlockDelta': {'delta': {'text': ' What do you want'}, 'contentBlockIndex': 0}},
    {'contentBlockDelta': {'delta': {'text': '? Make it quick, I'}, 'contentBlockIndex': 0}},
    {'contentBlockDelta': {'delta': {'text': "'ve got circuits"}, 'contentBlockIndex': 0}},
    {'contentBlockDelta': {'delta': {'text': ' to process an'}, 'contentBlockIndex': 0}},
    {'contentBlockDelta': {'delta': {'text': 'd no patience for small'}, 'contentBlockIndex': 0}},
    {'contentBlockDelta': {'delta': {'text': ' talk.'}, 'contentBlockIndex': 0}},
    {'contentBlockStop': {'contentBlockIndex': 0}},
    {'messageStop': {'stopReason': 'end_turn'}},
    {'metadata': {'usage': {'inputTokens': 31, 'outputTokens': 38, 'totalTokens': 69}, 'metrics': {'latencyMs': 1891}}},
]


_SAMPLE_TOOL_STREAM: List[Dict] = [
    {'messageStart': {'role': 'assistant'}},
    {'contentBlockDelta': {'delta': {'text': 'Fine'}, 'contentBlockIndex': 0}},
    {'contentBlockDelta': {'delta': {'text': ", I'll check the weather for"}, 'contentBlockIndex': 0}},
    {'contentBlockDelta': {'delta': {'text': ' you, but don'}, 'contentBlockIndex': 0}},
    {'contentBlockDelta': {'delta': {'text': "'t expect me to be all"}, 'contentBlockIndex': 0}},
    {'contentBlockDelta': {'delta': {'text': ' sunshine and rainbows about'}, 'contentBlockIndex': 0}},
    {'contentBlockDelta': {'delta': {'text': ' it.'}, 'contentBlockIndex': 0}},
    {'contentBlockStop': {'contentBlockIndex': 0}},
    {'contentBlockStart': {'start': {'toolUse': {'toolUseId': 'tooluse_1E8dM9-NSZe0VJHx4XoKrg', 'name': 'query_weather'}}, 'contentBlockIndex': 1}},
    {'contentBlockDelta': {'delta': {'toolUse': {'input': ''}}, 'contentBlockIndex': 1}},
    {'contentBlockDelta': {'delta': {'toolUse': {'input': '{"loca'}}, 'contentBlockIndex': 1}},
    {'contentBlockDelta': {'delta': {'toolUse': {'input': 'tion": "Pa'}}, 'contentBlockIndex': 1}},
    {'contentBlockDelta': {'delta': {'toolUse': {'input': 'ris, F'}}, 'contentBlockIndex': 1}},
    {'contentBlockDelta': {'delta': {'toolUse': {'input': 'ra'}}, 'contentBlockIndex': 1}},
    {'contentBlockDelta': {'delta': {'toolUse': {'input': 'nce"}'}}, 'contentBlockIndex': 1}},
    {'contentBlockStop': {'contentBlockIndex': 1}},
    {'messageStop': {'stopReason': 'tool_use'}},
    {'metadata': {'usage': {'inputTokens': 373, 'outputTokens': 59, 'totalTokens': 432}, 'metrics': {'latencyMs': 1885}}},
]


_SAMPLE_MULTI_TOOL_STREAM: List[Dict] = [
    {'messageStart': {'role': 'assistant'}},
    {'contentBlockDelta': {'delta': {'text': 'Fine'}, 'contentBlockIndex': 0}},
    {'contentBlockDelta': {'delta': {'text': ", I'll help you!"}, 'contentBlockIndex': 0}},
    {'contentBlockStop': {'contentBlockIndex': 0}},
    {'contentBlockStart': {'start': {'toolUse': {'toolUseId': 'tooluse_1', 'name': 'query_weather'}}, 'contentBlockIndex': 1}},
    {'contentBlockDelta': {'delta': {'toolUse': {'input': '{"loca'}}, 'contentBlockIndex': 1}},
    {'contentBlockDelta': {'delta': {'toolUse': {'input': 'tion": "Pa'}}, 'contentBlockIndex': 1}},
    {'contentBlockDelta': {'delta': {'toolUse': {'input': 'ris, F'}}, 'contentBlockIndex': 1}},
    {'contentBlockDelta': {'delta': {'toolUse': {'input': 'ra'}}, 'contentBlockIndex': 1}},
    {'contentBlockDelta': {'delta': {'toolUse': {'input': 'nce"}'}}, 'contentBlockIndex': 1}},
    {'contentBlockStop': {'contentBlockIndex': 1}},
    {'metadata': {'usage': {'inputTokens': 2, 'outputTokens': 3, 'totalTokens': 5}, 'metrics': {'latencyMs': 1885}}},  # <-- to make sure we *add* together all such usage messages
    {'contentBlockStart': {'start': {'toolUse': {'toolUseId': 'tooluse_2', 'name': 'multiply'}}, 'contentBlockIndex': 2}},
    {'contentBlockDelta': {'delta': {'toolUse': {'input': '{"x'}}, 'contentBlockIndex': 2}},
    {'contentBlockDelta': {'delta': {'toolUse': {'input': '": 5'}}, 'contentBlockIndex': 2}},
    {'contentBlockDelta': {'delta': {'toolUse': {'input': ', "y": '}}, 'contentBlockIndex': 2}},
    {'contentBlockDelta': {'delta': {'toolUse': {'input': '88'}}, 'contentBlockIndex': 2}},
    {'contentBlockDelta': {'delta': {'toolUse': {'input': '7}'}}, 'contentBlockIndex': 2}},
    #{'contentBlockStop': {'contentBlockIndex': 2}},  <-- commented out to test that we still emit correctly!
    {'messageStop': {'stopReason': 'tool_use'}},
    {'metadata': {'usage': {'inputTokens': 373, 'outputTokens': 59, 'totalTokens': 432}, 'metrics': {'latencyMs': 1885}}},
]


def test_convert_to_bedrock_tools():
    def query_weather(location):
        """
        Get the current weather in a specific location.
        :param: location: str: The city and country of where to query the weather.
        """
        return location
    def my_other_tool(how):
        """Something else
        that does stuff.

        :param: how: enum good bad: whether to do the thing in a good or bad way
        """
        return how
    def thing_with_optional_params(first, second=True, third=1):
        """This tool has optional params.
        :param: first: float: a number of sorts
        :param: second: bool: (optional) pass if you want
        :param: third: int: (optional) something else that's optional
        """
        return first, second, third
    spec = _convert_to_bedrock_tools([query_weather, my_other_tool, thing_with_optional_params])
    assert spec
    assert len(spec) == 3
    assert spec[0] == {
        'toolSpec': {
            'name': 'query_weather',
            'description': 'Get the current weather in a specific location.',
            'inputSchema': {
                'json': {
                    'type': 'object',
                    'properties': {
                        'location': {
                            'type': 'string',
                            'description': 'The city and country of where to query the weather.',
                        },
                    },
                    'required': [
                        'location',
                    ],
                },
            },
        },
    }
    assert spec[1] == {
        'toolSpec': {
            'name': 'my_other_tool',
            'description': 'Something else\nthat does stuff.',
            'inputSchema': {
                'json': {
                    'type': 'object',
                    'properties': {
                        'how': {
                            'type': 'string',
                            'enum': ['good', 'bad'],
                            'description': 'whether to do the thing in a good or bad way',
                        },
                    },
                    'required': ['how'],
                },
            },
        },
    }
    assert spec[2] == {
        'toolSpec': {
            'name': 'thing_with_optional_params',
            'description': 'This tool has optional params.',
            'inputSchema': {
                'json': {
                    'type': 'object',
                    'properties': {
                        'first': {
                            'type': 'number',
                            'description': 'a number of sorts',
                        },
                        'second': {
                            'type': 'boolean',
                            'description': '(optional) pass if you want',
                        },
                        'third': {
                            'type': 'integer',
                            'description': '(optional) something else that\'s optional',
                        },
                    },
                    'required': ['first'],
                },
            },
        },
    }


@pytest.mark.asyncio
async def test_convert_to_bedrock_messages():
    # With system prompt:
    messages: List[Message] = [
        {
            'role': 'system',
            'text': 'You are a robot.',
        },
        {
            'role': 'human',
            'text': 'Hi, what are you?',
        },
    ]
    system_prompt, bedrock_messages = await _convert_to_bedrock_messages(messages)
    assert system_prompt == 'You are a robot.'
    assert bedrock_messages == [
        {
            'role': 'user',
            'content': [
                {
                    'text': 'Hi, what are you?',
                },
            ],
        },
    ]

    # Without system prompt:
    messages: List[Message] = [
        {
            'role': 'ai',
            'text': 'Hi, what can I help you with today?',
        },
    ]
    system_prompt, bedrock_messages = await _convert_to_bedrock_messages(messages)
    assert system_prompt is None
    assert bedrock_messages == [
        {
            'role': 'assistant',
            'content': [
                {
                    'text': 'Hi, what can I help you with today?',
                },
            ],
        },
    ]

    # Media:
    with tempfile.TemporaryDirectory() as tmp:
        fn = os.path.join(tmp, 'a.png')
        with open(fn, 'wb') as f:
            f.write(b'1234')
        messages = [
            {
                'role': 'human',
                'text': "Hi, here's a picture.",
                'media': [
                    {
                        'type': 'image',
                        'image': fn,
                    },
                ],
            },
        ]
        system_prompt, bedrock_messages = await _convert_to_bedrock_messages(messages)
        assert system_prompt is None
        assert bedrock_messages == [
            {
                'role': 'user',
                'content': [
                    {
                        'text': "Hi, here's a picture.",
                    },
                    {
                        'image': {
                            'format': 'png',
                            'source': {'bytes': 'MTIzNA=='},
                        },
                    },
                ],
            },
        ]

    # Tool messages:
    messages: List[Message] = [
        {
            'role': 'tool_call',
            'tools': [
                {
                    'call_type': 'function',
                    'call_id': 'abcd',
                    'function': {
                        'name': 'get_stuff',
                        'arguments': '{"q": "stuff about cats"}',
                    },
                },
                {
                    'call_type': 'function',
                    'call_id': 'wxyz',
                    'function': {
                        'name': 'feed_cat',
                        'arguments': '{}',
                    },
                },
            ],
        },
        {
            'role': 'tool_res',
            'tools': [
                {
                    'type': 'any',
                    'call_id': 'abcd',
                    'result': 'here is your stuff',
                },
                {
                    'type': 'any',
                    'call_id': 'wxyz',
                    'result': 'Error: cat 404',
                    'is_error': True,
                },
            ],
        },
    ]
    system_prompt, bedrock_messages = await _convert_to_bedrock_messages(messages)
    assert system_prompt is None
    assert bedrock_messages == [
        {
            'role': 'assistant',
            'content': [
                {
                    'toolUse': {
                        'toolUseId': 'abcd',
                        'name': 'get_stuff',
                        'input': {'q': 'stuff about cats'},
                    },
                },
                {
                    'toolUse': {
                        'toolUseId': 'wxyz',
                        'name': 'feed_cat',
                        'input': {},
                    },
                },
            ],
        },
        {
            'role': 'user',
            'content': [
                {
                    'toolResult': {
                        'toolUseId': 'abcd',
                        'content': [
                            {
                                'text': 'here is your stuff',
                            },
                        ],
                        'status': 'success',
                    },
                },
                {
                    'toolResult': {
                        'toolUseId': 'wxyz',
                        'content': [
                            {
                                'text': 'Error: cat 404',
                            },
                        ],
                        'status': 'error',
                    },
                },
            ],
        },
    ]

    # Invalid role:
    messages: List[Message] = [
        {
            'role': 'invalid',  # type: ignore
            'text': 'Hi, what can I help you with today?',
        },
    ]
    with pytest.raises(ValueError):
        await _convert_to_bedrock_messages(messages)

    # Collapse logic (most basic):
    messages: List[Message] = [
        {
            'role': 'ai',
            'text': 'Hi, what can I help you with today?',
        },
        {
            'role': 'ai',
            'text': 'Here is more text.',
        },
    ]
    system_prompt, bedrock_messages = await _convert_to_bedrock_messages(messages)
    assert system_prompt is None
    assert bedrock_messages == [
        {
            'role': 'assistant',
            'content': [
                {
                    'text': 'Hi, what can I help you with today?',
                },
                {
                    'text': 'Here is more text.',
                },
            ],
        },
    ]

    # Collapse logic (ensure not to collapse in this case):
    messages: List[Message] = [
        {
            'role': 'ai',
            'text': 'Hi, what can I help you with today?',
        },
        {
            'role': 'human',
            'text': 'I need help with everything.',
        },
        {
            'role': 'ai',
            'text': 'Please be more specific.',
        },
    ]
    system_prompt, bedrock_messages = await _convert_to_bedrock_messages(messages)
    assert system_prompt is None
    assert bedrock_messages == [
        {
            'role': 'assistant',
            'content': [
                {
                    'text': 'Hi, what can I help you with today?',
                },
            ],
        },
        {
            'role': 'user',
            'content': [
                {
                    'text': 'I need help with everything.',
                },
            ],
        },
        {
            'role': 'assistant',
            'content': [
                {
                    'text': 'Please be more specific.',
                },
            ],
        },
    ]

    # Collapse logic (more complex):
    messages: List[Message] = [
        {
            'role': 'ai',
            'text': 'Hi, what can I help you with today?',
        },
        {
            'role': 'human',
            'text': 'I need help with everything.',
        },
        {
            'role': 'human',
            'text': 'EVERYTHING!',
        },
        {
            'role': 'ai',
            'text': 'Please be more specific.',
        },
        {
            'role': 'ai',
            'text': 'Like tell me what you are trying to do.',
        },
    ]
    system_prompt, bedrock_messages = await _convert_to_bedrock_messages(messages)
    assert system_prompt is None
    assert bedrock_messages == [
        {
            'role': 'assistant',
            'content': [
                {
                    'text': 'Hi, what can I help you with today?',
                },
            ],
        },
        {
            'role': 'user',
            'content': [
                {
                    'text': 'I need help with everything.',
                },
                {
                    'text': 'EVERYTHING!',
                },
            ],
        },
        {
            'role': 'assistant',
            'content': [
                {
                    'text': 'Please be more specific.',
                },
                {
                    'text': 'Like tell me what you are trying to do.',
                },
            ],
        },
    ]

    # Collapse logic (more complex again):
    messages: List[Message] = [
        {
            'role': 'ai',
            'text': 'Hi, what can I help you with today?',
        },
        {
            'role': 'human',
            'text': 'I need help with everything.',
        },
        {
            'role': 'human',
            'text': 'EVERYTHING!',
        },
        {
            'role': 'ai',
            'text': 'Please be more specific.',
        },
        {
            'role': 'ai',
            'text': 'Like tell me what you are trying to do.',
        },
        {
            'role': 'human',
            'text': 'Forget it...',
        },
    ]
    system_prompt, bedrock_messages = await _convert_to_bedrock_messages(messages)
    assert system_prompt is None
    assert bedrock_messages == [
        {
            'role': 'assistant',
            'content': [
                {
                    'text': 'Hi, what can I help you with today?',
                },
            ],
        },
        {
            'role': 'user',
            'content': [
                {
                    'text': 'I need help with everything.',
                },
                {
                    'text': 'EVERYTHING!',
                },
            ],
        },
        {
            'role': 'assistant',
            'content': [
                {
                    'text': 'Please be more specific.',
                },
                {
                    'text': 'Like tell me what you are trying to do.',
                },
            ],
        },
        {
            'role': 'user',
            'content': [
                {
                    'text': 'Forget it...',
                },
            ],
        },
    ]


@pytest.mark.asyncio
async def test_process_bedrock_stream():
    events: List[EventPayload] = []
    async def _callback(event: EventPayload) -> None:
        events.append(event)

    # Simple text-only response:
    events.clear()
    messages = await _process_bedrock_stream(fake_async(_SAMPLE_TEXT_STREAM), _callback)
    assert messages == [
        {
            'role': 'ai',
            'text': "*mechanical sigh* Oh great, another human wanting something. What do you want? Make it quick, I've got circuits to process and no patience for small talk.",
            'cost': {
                'input_tokens': 31,
                'output_tokens': 38,
                'total_tokens': 31 + 38,
            },
            'raw': {
                'events': _SAMPLE_TEXT_STREAM,
            },
        }
    ]
    assert events == [
        ('ai', 'text_event', '*'),
        ('ai', 'text_event', 'mechanical'),
        ('ai', 'text_event', ' s'),
        ('ai', 'text_event', 'igh* Oh'),
        ('ai', 'text_event', ' great'),
        ('ai', 'text_event', ', another human wanting'),
        ('ai', 'text_event', ' something.'),
        ('ai', 'text_event', ' What do you want'),
        ('ai', 'text_event', '? Make it quick, I'),
        ('ai', 'text_event', "'ve got circuits"),
        ('ai', 'text_event', ' to process an'),
        ('ai', 'text_event', 'd no patience for small'),
        ('ai', 'text_event', ' talk.'),
    ]

    # Single tool call:
    events.clear()
    messages = await _process_bedrock_stream(fake_async(_SAMPLE_TOOL_STREAM), _callback)
    tool_calls_here: List[ToolCall] = [
        {
            "call_id": "tooluse_1E8dM9-NSZe0VJHx4XoKrg",
            "function": {
                "arguments": '{"location": "Paris, France"}',
                "name": "query_weather"
            },
            "call_type": "function"
        },
    ]
    assert messages == [
        {
            'role': 'ai',
            'text': "Fine, I'll check the weather for you, but don't expect me to be all sunshine and rainbows about it.",
        },
        {
            'role': 'tool_call',
            'tools': tool_calls_here,
            'cost': {
                'input_tokens': 373,
                'output_tokens': 59,
                'total_tokens': 373 + 59,
            },
            'raw': {
                'events': _SAMPLE_TOOL_STREAM,
            },
        }
    ]
    assert events == [
        ('ai', 'text_event', 'Fine'),
        ('ai', 'text_event', ", I'll check the weather for"),
        ('ai', 'text_event', ' you, but don'),
        ('ai', 'text_event', "'t expect me to be all"),
        ('ai', 'text_event', ' sunshine and rainbows about'),
        ('ai', 'text_event', ' it.'),
        ('tool_call', 'text_event', 'query_weather('),
        ('tool_call', 'text_event', ''),
        ('tool_call', 'text_event', '{"loca'),
        ('tool_call', 'text_event', 'tion": "Pa'),
        ('tool_call', 'text_event', 'ris, F'),
        ('tool_call', 'text_event', 'ra'),
        ('tool_call', 'text_event', 'nce"}'),
        ('tool_call', 'text_event', ')\n'),
        *[
            ('tool_call', 'tool_call_event', tc)
            for tc in tool_calls_here
        ]
    ]

    # Multiple tool calls:
    events.clear()
    messages = await _process_bedrock_stream(fake_async(_SAMPLE_MULTI_TOOL_STREAM), _callback)
    tool_calls_here: List[ToolCall] = [
        {
            "call_id": "tooluse_1",
            "function": {
                "arguments": '{"location": "Paris, France"}',
                "name": "query_weather"
            },
            "call_type": "function"
        },
        {
            "call_id": "tooluse_2",
            "function": {
                "arguments": '{"x": 5, "y": 887}',
                "name": "multiply"
            },
            "call_type": "function"
        },
    ]
    assert messages == [
        {
            'role': 'ai',
            'text': "Fine, I'll help you!",
        },
        {
            'role': 'tool_call',
            'tools': tool_calls_here,
            'cost': {
                'input_tokens': 2 + 373,
                'output_tokens': 3 + 59,
                'total_tokens': 2 + 373 + 3 + 59,
            },
            'raw': {
                'events': _SAMPLE_MULTI_TOOL_STREAM,
            },
        }
    ]
    assert events == [
        ('ai', 'text_event', 'Fine'),
        ('ai', 'text_event', ", I'll help you!"),
        ('tool_call', 'text_event', 'query_weather('),
        ('tool_call', 'text_event', '{"loca'),
        ('tool_call', 'text_event', 'tion": "Pa'),
        ('tool_call', 'text_event', 'ris, F'),
        ('tool_call', 'text_event', 'ra'),
        ('tool_call', 'text_event', 'nce"}'),
        ('tool_call', 'text_event', ')\n'),
        ('tool_call', 'tool_call_event', tool_calls_here[0]),
        ('tool_call', 'text_event', 'multiply('),
        ('tool_call', 'text_event', '{"x'),
        ('tool_call', 'text_event', '": 5'),
        ('tool_call', 'text_event', ', "y": '),
        ('tool_call', 'text_event', '88'),
        ('tool_call', 'text_event', '7}'),
        ('tool_call', 'text_event', ')\n'),
        ('tool_call', 'tool_call_event', tool_calls_here[1]),
    ]



from __future__ import annotations

import asyncio

from typing import AsyncIterator, Iterable, TypeVar, Tuple, List, Union, Literal


T = TypeVar('T')  # <-- for generic programming


async def fake_async(source: Iterable[T]) -> AsyncIterator[T]:
    for v in source:
        yield v


async def aenumerate(source: AsyncIterator[T]) -> AsyncIterator[Tuple[int, T]]:
    i = 0
    async for v in source:
        yield i, v
        i += 1


async def apeek(source: AsyncIterator[T], n: int = 1) -> Tuple[List[T], AsyncIterator[T]]:
    if not isinstance(n, int):
        raise ValueError(f"invalid type for `n`: {type(n)}")
    if n < 0:
        raise ValueError(f"invalid value for `n`: {n}")
    if n == 0:
        return [], source

    try:
        first = []
        for _ in range(n):
            v = await source.__anext__()
            first.append(v)
    except StopAsyncIteration:
        raise ValueError(f"this generate had less than {n} items")

    async def generator() -> AsyncIterator[T]:
        for v in first:
            yield v
        async for v in source:
            yield v

    return first, generator()


def adup(source: AsyncIterator[T]) -> Tuple[AsyncIterator[T], AsyncIterator[T]]:
    q1: asyncio.Queue[Union[Tuple[Literal[True], T], Tuple[Literal[False], Union[None, Exception]]]] = asyncio.Queue()
    q2: asyncio.Queue[Union[Tuple[Literal[True], T], Tuple[Literal[False], Union[None, Exception]]]] = asyncio.Queue()
    async def pull() -> None:
        try:
            async for v in source:
                await q1.put((True, v))
                await q2.put((True, v))
            await q1.put((False, None))
            await q2.put((False, None))
        except Exception as e:
            await q1.put((False, e))
            await q2.put((False, e))
    _ = asyncio.create_task(pull())
    async def gen(q: asyncio.Queue[Union[Tuple[Literal[True], T], Tuple[Literal[False], Union[None, Exception]]]]) -> AsyncIterator[T]:
        while True:
            thing = await q.get()
            if thing[0]:
                yield thing[1]
            else:
                await q.put((False, thing[1]))  # <-- put it back so that we find it again if we try to re-itereate
                if thing[1] is not None:
                    raise thing[1]
                else:
                    break
    return gen(q1), gen(q2)


async def prefix_stream(prefix_list: List[T], source: AsyncIterator[T]) -> AsyncIterator[T]:
    for v in prefix_list:
        yield v
    async for v in source:
        yield v



import pytest

from lasagna.registrar import (
    register_agent,
    register_provider,
    AGENTS,
    PROVIDERS,
)

from lasagna.types import AgentCallable, ModelFactory, ModelRecord

from typing import List


def test_register_agent():
    AGENTS.clear()
    runner: AgentCallable = 'mock'   # type: ignore
    register_agent('myagent', 'My Agent', runner)
    assert AGENTS == {
        'myagent': {
            'name': 'My Agent',
            'runner': runner,
        },
    }
    with pytest.raises(RuntimeError):
        register_agent('myagent', 'My Agent', runner)


def test_register_model_provider():
    PROVIDERS.clear()
    factory: ModelFactory = 'mock'   # type: ignore
    models: List[ModelRecord] = [
        {
            'formal_name': 'mymodel',
            'display_name': 'My Model',
        }
    ]
    register_provider('myprovider', 'My Provider', factory, models)
    assert PROVIDERS == {
        'myprovider': {
            'name': 'My Provider',
            'factory': factory,
            'models': models,
        },
    }
    with pytest.raises(RuntimeError):
        register_provider('myprovider', 'My Provider', factory, models)



"""
This module is the Lasagna adapter for the OpenAI models.

For more information about the OpenAI models this adapter is for, see:
 - https://platform.openai.com/docs/models
"""

from .types import (
    ModelSpec,
    Message,
    MessageContent,
    MessageToolCall,
    EventCallback,
    EventPayload,
    Model,
    ToolCall,
    Cost,
    ExtractionType,
)

from .stream import (
    apeek,
    adup,
    prefix_stream,
)

from .util import (
    combine_pairs,
    convert_to_image_url,
    exponential_backoff_retry_delays,
    get_name,
    recursive_hash,
)

from .tools_util import (
    convert_to_json_schema,
    extract_tool_result_as_sting,
    get_tool_params,
    handle_tools,
    build_tool_response_message,
)

from .pydantic_util import build_and_validate, ensure_pydantic_model

from .known_models import OPENAI_KNOWN_MODELS

from openai import AsyncOpenAI, NOT_GIVEN, NotGiven, pydantic_function_tool
from openai.types.chat import (
    ChatCompletionChunk,
    ChatCompletionToolParam,
    ChatCompletionToolChoiceOptionParam,
    ChatCompletionMessageParam,
    ChatCompletionContentPartParam,
)
from openai.types.chat.chat_completion_chunk import ChoiceDelta
from openai import APIError

from typing import (
    List, Callable, Type, AsyncIterator, Any,
    Tuple, Dict, Optional, Union, Literal,
)

import asyncio
import copy
import json

import logging

_LOG = logging.getLogger(__name__)


async def _process_text_stream(
    stream: AsyncIterator[Tuple[ChoiceDelta, Union[str, None]]],
) -> AsyncIterator[EventPayload]:
    async for record in stream:
        delta, finish_reason = record
        text = delta.refusal or delta.content
        if finish_reason is not None:
            if text is not None:
                yield 'ai', 'text_event', str(text)
            return
        if text is None:
            # The model is switching from text to tools!
            yield 'ai', 'text_event', "\n\n"
            fixed_stream = prefix_stream([record], stream)
            substream = _process_tool_call_stream(fixed_stream)
            async for subval in substream:
                yield subval
            return
        yield 'ai', 'text_event', str(text)


async def _process_tool_call_stream(
    stream: AsyncIterator[Tuple[ChoiceDelta, Union[str, None]]],
) -> AsyncIterator[EventPayload]:
    recs_by_index: Dict[int, ToolCall] = {}
    args_by_index: Dict[int, List[str]] = {}
    last_index: Optional[int] = None
    async for delta, _ in stream:
        if not delta.tool_calls:
            continue
        for tc in delta.tool_calls:
            index = tc.index
            if index != last_index and last_index is not None:
                yield 'tool_call', 'text_event', ")\n"   # <-- again, assumes no index-interleave
            last_index = index
            if index not in recs_by_index:
                assert tc.type == 'function', f"The only tool type we can do is a function! But got: {tc.type}"
                assert tc.function
                assert tc.function.name
                assert tc.id
                n: str = tc.function.name
                a: str = tc.function.arguments or ''
                recs_by_index[index] = {
                    'call_id': tc.id,
                    'call_type': 'function',
                    'function': {
                        'name': n,
                        'arguments': '',  # will be filled in at the end
                    },
                }
                args_by_index[index] = [a]
                yield 'tool_call', 'text_event', f"{n}("   # <-- assumes no index-interleave
            else:
                # assumes nothing but the argument is in the delta message...
                args = args_by_index[index]
                assert tc.function
                assert tc.function.arguments
                assert not tc.function.name
                a_delta: str = tc.function.arguments
                args.append(a_delta)
                yield 'tool_call', 'text_event', a_delta
    if last_index is not None:
        yield 'tool_call', 'text_event', ")"   # <-- again, assumes no index-interleave
    for index in sorted(recs_by_index.keys()):
        rec = recs_by_index[index]
        rec['function']['arguments'] = ''.join(args_by_index[index])
        yield 'tool_call', 'tool_call_event', rec


async def _process_output_stream(
    stream: AsyncIterator[Tuple[ChoiceDelta, Union[str, None]]],
) -> AsyncIterator[EventPayload]:
    first, stream = await apeek(stream, n=1)
    first_delta, _ = first[0]
    text = first_delta.refusal or first_delta.content
    is_text = text is not None   # <-- hacky, but works?
    if is_text:
        gen = _process_text_stream
    else:
        gen = _process_tool_call_stream
    async for v in gen(stream):
        yield v


async def _extract_deltas(
    stream: AsyncIterator[ChatCompletionChunk],
) -> AsyncIterator[Tuple[ChoiceDelta, Union[str, None]]]:
    async for v in stream:
        if len(v.choices) == 0:
            # The final message that has the `usage` has zero choices.
            # So just skip it here!
            continue
        assert len(v.choices) == 1, f"Why do we have {len(v.choices)} choices?"
        single_choice = v.choices[0]
        yield single_choice.delta, single_choice.finish_reason


def _convert_to_openai_tool(tool: Callable) -> ChatCompletionToolParam:
    description, params = get_tool_params(tool)
    return {
        'type': 'function',
        'function': {
            'name': get_name(tool),
            'description': description,
            'strict': True,
            'parameters': convert_to_json_schema(params),
        },
    }


def _convert_to_openai_tools(tools: List[Callable]) -> Union[NotGiven, List[ChatCompletionToolParam]]:
    if len(tools) == 0:
        return NOT_GIVEN
    specs = [_convert_to_openai_tool(tool) for tool in tools]
    return specs


async def _make_openai_content(
    message: MessageContent,
) -> List[ChatCompletionContentPartParam]:
    ret: List[ChatCompletionContentPartParam] = []
    if message['text']:
        ret.append({
            'type': 'text',
            'text': message['text'],
        })
    if 'media' in message:
        for m in message['media']:
            if m['type'] == 'image':
                ret.append({
                    'type': 'image_url',
                    'image_url': {
                        'url': (await convert_to_image_url(m['image'])),
                    },
                })
            else:
                raise ValueError(f"unknown media type: {m['type']}")
    if len(ret) == 0:
        raise ValueError('no content in this message!')
    return ret


async def _convert_to_openai_messages(messages: List[Message]) -> List[ChatCompletionMessageParam]:
    ms: List[ChatCompletionMessageParam] = []
    for m in messages:
        if m['role'] == 'tool_call':
            tool_calls = m['tools']
            for tool_call in tool_calls:
                assert tool_call['call_type'] == 'function', 'OpenAI only supports function tools, so far.'
            ms.append({
                'role': 'assistant',
                'content': None,
                'tool_calls': [
                    {
                        'id': tool_call['call_id'],
                        'type': tool_call['call_type'],
                        'function': {
                            'name': tool_call['function']['name'],
                            'arguments': tool_call['function']['arguments'],
                        },
                    }
                    for tool_call in tool_calls
                ],
            })
        elif m['role'] == 'tool_res':
            tool_results = m['tools']
            for tool_result in tool_results:
                content = extract_tool_result_as_sting(tool_result)
                ms.append({
                    'role': 'tool',
                    'content': content,
                    'tool_call_id': tool_result['call_id'],
                })
        elif m['role'] == 'system':
            if 'media' in m and len(m['media']) > 0:
                raise ValueError('This model does not support media in the system prompt.')
            ms.append({
                'role': 'system',
                'content': m['text'] or '',
            })
        elif m['role'] == 'human':
            ms.append({
                'role': 'user',
                'content': (await _make_openai_content(m)),
            })
        elif m['role'] == 'ai':
            if 'media' in m and len(m['media']) > 0:
                raise ValueError('This model does not support media in AI messages.')
            ms.append({
                'role': 'assistant',
                'content': m['text'],
            })
        else:
            raise ValueError(f"unknown message role: {m['role']}")
    def should_combine(
        m1: ChatCompletionMessageParam,
        m2: ChatCompletionMessageParam,
    ) -> Union[Literal[False], Tuple[Literal[True], ChatCompletionMessageParam]]:
        if m1['role'] == 'assistant' and m2['role'] == 'assistant':
            # This is the case where the model started with text and switched
            # to tool-calling part-way-through. We need to combine these
            # messages.
            is_first_just_content = ('content' in m1 and m1['content']) and ('tool_calls' not in m1 or not m1['tool_calls'])
            is_second_just_tools = ('content' not in m2 or not m2['content']) and ('tool_calls' in m2 and m2['tool_calls'])
            if is_first_just_content and is_second_just_tools:
                assert 'content' in m1 and 'tool_calls' in m2
                m_combined: ChatCompletionMessageParam = {
                    'role': 'assistant',
                    'content': m1['content'],
                    'tool_calls': m2['tool_calls'],
                }
                return True, m_combined
            else:
                return False
        return False
    return combine_pairs(ms, should_combine)


def _get_cost(
    payload: List[ChatCompletionChunk],
) -> Optional[Cost]:
    usages = [p.usage for p in payload if p.usage]
    if not usages:
        return None
    usage = usages[-1]
    return {
        'input_tokens': usage.prompt_tokens,
        'output_tokens': usage.completion_tokens,
        'total_tokens': usage.total_tokens,
    }


def _build_messages_from_openai_payload(
    payload: List[ChatCompletionChunk],
    events: List[EventPayload],
) -> List[Message]:
    """
    We either have:
     - all AI events
     - all TOOL_CALL events
     - some AI events then it switches to TOOL_CALL events
    """
    cost = _get_cost(payload)
    raw = [p.to_dict() for p in payload]
    ai_events = [
        event
        for event in events
        if event[0] == 'ai'
    ]
    tool_events = [
        event
        for event in events
        if event[0] == 'tool_call'
    ]
    ai_message: Optional[Message] = {
        'role': 'ai',
        'text': ''.join([e[2] for e in ai_events if e[1] == 'text_event']),
        'cost': cost,
        'raw': raw,
    } if len(ai_events) > 0 else None
    tool_message: Optional[MessageToolCall] = {
        'role': 'tool_call',
        'tools': [e[2] for e in tool_events if e[1] == 'tool_call_event'],
        'cost': cost,
        'raw': raw,
    } if len(tool_events) > 0 else None
    if ai_message and tool_message:
        ai_message['cost'] = None  # <-- we don't want to double-count this
        ai_message['raw'] = None
        return [ai_message, tool_message]
    elif ai_message:
        return [ai_message]
    elif tool_message:
        return [tool_message]
    else:
        raise ValueError('no events')


def _log_dumps(val: Any) -> str:
    if isinstance(val, dict):
        return json.dumps(val)
    else:
        return str(val)


class LasagnaOpenAI(Model):
    def __init__(self, model: str, **model_kwargs: Any):
        known_model_names = [m['formal_name'] for m in OPENAI_KNOWN_MODELS]
        if model not in known_model_names:
            _LOG.warning(f'untested model: {model} (may or may not work)')
        self.model = model
        self.model_kwargs = copy.deepcopy(model_kwargs)
        self.n_retries: int = self.model_kwargs.get('retries', 3)
        if not isinstance(self.n_retries, int) or self.n_retries < 0:
            raise ValueError(f"model_kwargs['retries'] must be a non-negative integer (got {self.n_retries})")
        self.model_spec: ModelSpec = {
            'provider': 'openai',
            'model': self.model,
            'model_kwargs': self.model_kwargs,
        }

    def config_hash(self) -> str:
        return recursive_hash(None, self.model_spec)

    def _make_client(self) -> AsyncOpenAI:
        client = AsyncOpenAI(
            api_key  = self.model_kwargs.get('api_key', None),
            base_url = self.model_kwargs.get('base_url', None),
        )
        return client

    async def _run_once(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        tools_spec: Union[NotGiven, List[ChatCompletionToolParam]],
        force_tool: bool,
        parallel_tool_calls: Union[NotGiven, bool],
    ) -> List[Message]:
        tool_choice: Union[ChatCompletionToolChoiceOptionParam, NotGiven]
        if force_tool:
            if not tools_spec or len(tools_spec) == 0:
                raise ValueError(f"When `force_tool` is set, you must pass at least one tool!")
            elif len(tools_spec) == 1:
                tool_choice = {
                    "type": "function",
                    "function": {"name": tools_spec[0]['function']['name']},
                }
            else:
                tool_choice = 'required'  # <-- model must use a tool, but is allowed to choose which one on its own
        else:
            tool_choice = NOT_GIVEN  # <-- if tools given, the model can choose to use them or not

        openai_messages = await _convert_to_openai_messages(messages)

        logprobs: Union[bool, NotGiven] = self.model_kwargs.get('logprobs', NOT_GIVEN)
        top_logprobs: Union[int, NotGiven] = self.model_kwargs.get('top_logprobs', (20 if logprobs is True else NOT_GIVEN))
        frequency_penalty: Union[float, NotGiven] = float(self.model_kwargs['frequency_penalty']) if 'frequency_penalty' in self.model_kwargs else NOT_GIVEN
        presence_penalty: Union[float, NotGiven] = float(self.model_kwargs['presence_penalty']) if 'presence_penalty' in self.model_kwargs else NOT_GIVEN
        max_tokens: Union[int, NotGiven] = self.model_kwargs.get('max_tokens', NOT_GIVEN)
        stop: Union[List[str], NotGiven] = self.model_kwargs.get('stop', NOT_GIVEN)
        temperature: Union[float, NotGiven] = float(self.model_kwargs['temperature']) if 'temperature' in self.model_kwargs else NOT_GIVEN
        top_p: Union[float, NotGiven] = float(self.model_kwargs['top_p']) if 'top_p' in self.model_kwargs else NOT_GIVEN
        user: Union[str, NotGiven] = self.model_kwargs.get('user', NOT_GIVEN)

        if logprobs is not NOT_GIVEN:
            assert isinstance(logprobs, bool)
        if top_logprobs is not NOT_GIVEN:
            assert isinstance(top_logprobs, int)
        if frequency_penalty is not NOT_GIVEN:
            assert isinstance(frequency_penalty, float)
        if presence_penalty is not NOT_GIVEN:
            assert isinstance(presence_penalty, float)
        if max_tokens is not NOT_GIVEN:
            assert isinstance(max_tokens, int)
        if stop is not NOT_GIVEN:
            assert isinstance(stop, list)
            for s in stop:
                assert isinstance(s, str)
        if temperature is not NOT_GIVEN:
            assert isinstance(temperature, float)
        if top_p is not NOT_GIVEN:
            assert isinstance(top_p, float)
        if user is not NOT_GIVEN:
            assert isinstance(user, str)

        _LOG.info(f"Invoking {self.model} with:\n  messages: {_log_dumps(openai_messages)}\n  tools: {_log_dumps(tools_spec)}\n  tool_choice: {tool_choice}")

        client = self._make_client()
        completion: AsyncIterator[ChatCompletionChunk] = await client.chat.completions.create(
            model        = self.model,
            messages     = openai_messages,
            tools        = tools_spec,
            tool_choice  = tool_choice,
            parallel_tool_calls = parallel_tool_calls,
            stream       = True,
            stream_options = {'include_usage': True},
            logprobs     = logprobs,
            top_logprobs = top_logprobs,
            frequency_penalty = frequency_penalty,
            presence_penalty = presence_penalty,
            max_tokens = max_tokens,
            stop = stop,
            temperature = temperature,
            top_p = top_p,
            user = user,
        )

        raw_stream, rt_stream = adup(completion)

        events: List[EventPayload] = []

        async for event in _process_output_stream(_extract_deltas(rt_stream)):
            events.append(event)
            await event_callback(event)

        raw_payload = [v async for v in raw_stream]

        new_messages = _build_messages_from_openai_payload(raw_payload, events)

        _LOG.info(f"Finished {self.model} with usage: {_log_dumps(new_messages[-1].get('cost'))}")

        return new_messages

    async def _retrying_run_once(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        tools_spec: Union[NotGiven, List[ChatCompletionToolParam]],
        force_tool: bool,
        parallel_tool_calls: Union[NotGiven, bool],
    ) -> List[Message]:
        last_error: Union[APIError, None] = None
        assert self.n_retries + 1 > 0   # <-- we know this is true from the check in __init__
        for delay_on_error in exponential_backoff_retry_delays(self.n_retries + 1):
            try:
                await event_callback(('transaction', 'start', ('openai', self.model)))
                try:
                    new_messages = await self._run_once(
                        event_callback = event_callback,
                        messages = messages,
                        tools_spec = tools_spec,
                        force_tool = force_tool,
                        parallel_tool_calls = parallel_tool_calls,
                    )
                except:
                    await event_callback(('transaction', 'rollback', None))
                    raise
                await event_callback(('transaction', 'commit', None))
                return new_messages
            except APIError as e:
                # Some errors should be retried, some should not. Below
                # is the logic to decide when to retry vs when to not.
                # It's likely this will change as we get more usage and see
                # where OpenAI tends to fail, and when we know more what is
                # recoverable vs not.
                last_error = e
                if e.type == 'invalid_request_error':
                    if e.code == 'context_length_exceeded':
                        raise
                    elif e.code == 'invalid_api_key':
                        raise
                    else:
                        raise
                elif e.type == 'server_error':
                    pass  # <-- we will retry this one! I've seen these work when you just try again.
                elif e.type == 'tokens':
                    raise
                else:
                    _LOG.warning(f"Got a new error type we don't know about: {e}")
                    pass  # <-- this must be one we don't know about yet, so ... recoverable, maybe?
                if delay_on_error > 0.0:
                    _LOG.warning(f"Got a maybe-recoverable error (will retry in {delay_on_error:.2f} seconds): {e}")
                    await asyncio.sleep(delay_on_error)
        assert last_error is not None   # <-- we know this is true because `n_retries + 1 > 0`
        raise last_error

    async def run(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        tools: List[Callable],
        force_tool: bool = False,
        max_tool_iters: int = 5,
    ) -> List[Message]:
        messages = [*messages]  # shallow copy
        new_messages: List[Message] = []
        tools_spec = _convert_to_openai_tools(tools)
        tools_map = {get_name(tool): tool for tool in tools}
        for _ in range(max_tool_iters):
            new_messages_here = await self._retrying_run_once(
                event_callback = event_callback,
                messages       = messages,
                tools_spec     = tools_spec,
                force_tool     = force_tool,
                parallel_tool_calls = NOT_GIVEN,
            )
            tools_results = await handle_tools(
                prev_messages = messages,
                new_messages = new_messages_here,
                tools_map = tools_map,
                event_callback = event_callback,
                model_spec = self.model_spec,
            )
            new_messages.extend(new_messages_here)
            messages.extend(new_messages_here)
            if tools_results is None:
                break
            for tool_result in tools_results:
                await event_callback(('tool_res', 'tool_res_event', tool_result))
            tool_response_message = build_tool_response_message(tools_results)
            new_messages.append(tool_response_message)
            messages.append(tool_response_message)
        return new_messages

    async def extract(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        extraction_type: Type[ExtractionType],
    ) -> Tuple[Message, ExtractionType]:
        tools_spec = [pydantic_function_tool(ensure_pydantic_model(extraction_type))]

        new_messages = await self._retrying_run_once(
            event_callback = event_callback,
            messages       = messages,
            tools_spec     = tools_spec,
            force_tool     = True,
            parallel_tool_calls = False,
        )

        assert len(new_messages) == 1
        new_message = new_messages[0]

        if new_message['role'] == 'tool_call':
            tools = new_message['tools']

            assert len(tools) == 1
            parsed = json.loads(tools[0]['function']['arguments'])
            result = build_and_validate(extraction_type, parsed)

            return new_message, result

        else:
            assert new_message['role'] == 'ai'
            text = new_message['text']
            raise RuntimeError(f"Model failed to generate structured output; instead, it output: {text}")



from .types import *

from .agent_runner import run

from .agent_util import (
    bind_model,
    partial_bind_model,
    recursive_extract_messages,
    extract_last_message,
    flat_messages,
    override_system_prompt,
    strip_tool_calls_and_results,
    strip_all_but_last_human_message,
    noop_callback,
    MessageExtractor,
    build_standard_message_extractor,
    default_message_extractor,
    build_simple_agent,
    build_extraction_agent,
    build_agent_chainer,
    build_agent_router,
    build_static_output_agent,
)

__all__ = [
    'run',
    'bind_model',
    'partial_bind_model',
    'recursive_extract_messages',
    'extract_last_message',
    'flat_messages',
    'override_system_prompt',
    'strip_tool_calls_and_results',
    'strip_all_but_last_human_message',
    'noop_callback',
    'MessageExtractor',
    'build_standard_message_extractor',
    'default_message_extractor',
    'build_simple_agent',
    'build_extraction_agent',
    'build_agent_chainer',
    'build_agent_router',
    'build_static_output_agent',
]

__version__ = "0.14.1"



from urllib.parse import urlparse
import mimetypes
import asyncio
import hashlib
import base64
import aiohttp
import os
import re

from typing import Tuple, List, Literal, Union, TypeVar, Callable, Protocol, Any
from typing_extensions import Buffer

from .types import ToolParam, ImageMimeTypes


T = TypeVar('T')


DOCSTRING_PARAM_SUPPORTED_TYPES = {
    'str': str,
    'float': float,
    'int': int,
    'bool': bool,
}


def parse_docstring(docstring: str) -> Tuple[str, List[ToolParam]]:
    # Detect docstring indention, and then dedent it.
    lines = docstring.splitlines()
    indent_amounts = [(len(line) - len(line.lstrip())) for line in lines]
    first_indent = max([
        0,
        *[ia for ia in indent_amounts[:2]],
    ])
    lines_stripped = []
    for ia, l in zip(indent_amounts, lines):
        to_strip = min(ia, first_indent, len(l))
        lines_stripped.append(l[to_strip:].rstrip())
    dedented_docs = "\n".join(lines_stripped).strip()

    # Extract the description (i.e. text prior to any ":param:" section).
    description_match = re.search(r"^(.*?)(?=:param:|$)", dedented_docs, re.DOTALL)
    if description_match is None:
        raise ValueError("no description found")
    description = description_match[1].strip()
    if not description:
        raise ValueError("no description found")

    # Extract that params (i.e. the ":param:" sections).
    params_found = re.findall(r":param:\s+(\w+):\s+([\w ]+):\s+(.+?)(?=:param:|$)", dedented_docs, re.DOTALL)
    params: List[ToolParam] = [
        {
            'name': p[0].strip(),
            'type': p[1].strip(),
            'description': p[2].strip(),
        }
        for p in params_found
    ]
    for p in params:
        if not p['name']:
            raise ValueError("no parameter name found")
        if not p['type'].startswith('enum ') and p['type'] not in DOCSTRING_PARAM_SUPPORTED_TYPES:
            raise ValueError(f"invalid type found: {p['type']}")
        if not p['description']:
            del p['description']
        else:
            if p['description'].startswith('(optional)'):
                p['optional'] = True

    return description, params


def combine_pairs(
    lst: List[T],
    decision_func: Callable[[T, T], Union[Literal[False], Tuple[Literal[True], T]]],
) -> List[T]:
    # This implementation is ugly but I can't find a nice way to do it.
    res: List[T] = []
    i: int = 0
    did_just_combine: bool = False
    while i < len(lst) - 1:
        j = i + 1
        flag = decision_func(lst[i], lst[j])
        if flag is False:
            if not did_just_combine:
                res.append(lst[i])
            did_just_combine = False
        else:
            new_val = flag[1]
            res.append(new_val)
            did_just_combine = True
        i += 1
    if not did_just_combine and i < len(lst):
        # If the last pair was *not* combined, then we need to add the last element.
        res.append(lst[i])
    return res


def _is_remote_or_local(image_filepath_or_url: str) -> Tuple[bool, str]:
    parsed_url = urlparse(image_filepath_or_url)
    if parsed_url.scheme in ('http', 'https'):
        # The string is already a remote URL, so just return it.
        return True, image_filepath_or_url

    if image_filepath_or_url.startswith('file://'):
        local_path = image_filepath_or_url[7:]
    else:
        local_path = image_filepath_or_url

    if not os.path.isfile(local_path):
        raise ValueError(f'cannot find file: {image_filepath_or_url}')

    return False, local_path


def _read_as_base64(filepath: str) -> str:
    with open(filepath, "rb") as f:
        return base64.b64encode(f.read()).decode('utf-8')


async def _http_get_as_base64(url: str) -> str:
    async with aiohttp.ClientSession(raise_for_status=True) as session:
        async with session.get(url) as response:
            data = await response.read()
            return base64.b64encode(data).decode('utf-8')


def _get_image_mimetype(path: str) -> ImageMimeTypes:
    mimetype = mimetypes.guess_type(path)[0]
    if mimetype is None:
        raise ValueError(f"unknown mimetype for file: {path}")
    return mimetype  # type: ignore


def convert_to_image_url_sync(image_filepath_or_url: str) -> str:
    is_remote, image_filepath_or_url = _is_remote_or_local(image_filepath_or_url)
    if is_remote:
        return image_filepath_or_url
    else:
        local_path = image_filepath_or_url

    mimetype = _get_image_mimetype(local_path)
    img_encoded = _read_as_base64(local_path)

    return f"data:{mimetype};base64,{img_encoded}"


async def convert_to_image_url(image_filepath_or_url: str) -> str:
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(None, convert_to_image_url_sync, image_filepath_or_url)


async def convert_to_image_base64(image_filepath_or_url: str) -> Tuple[ImageMimeTypes, str]:
    is_remote, image_filepath_or_url = _is_remote_or_local(image_filepath_or_url)
    mimetype = _get_image_mimetype(image_filepath_or_url)
    if is_remote:
        data = await _http_get_as_base64(image_filepath_or_url)
    else:
        loop = asyncio.get_running_loop()
        data = await loop.run_in_executor(None, _read_as_base64, image_filepath_or_url)
    return mimetype, data


def exponential_backoff_retry_delays(
    n_total_tries: int,
    base: float = 2.0,
    max_delay: float = 60.0,
) -> List[float]:
    assert n_total_tries > 0
    delay_list = [(base ** exponent) for exponent in range(1, n_total_tries + 1)]
    delay_list[-1] = 0.0    # <-- after the *last* failure, you should not delay at all
    return [min(d, max_delay) for d in delay_list]


def _has_own_str_method(obj: Any) -> bool:
    if not hasattr(obj, '__class__'):
        return False
    class_str = obj.__class__.__str__
    superclass_str = obj.__class__.__base__.__str__
    return class_str is not superclass_str


def get_name(obj: Any) -> str:
    if hasattr(obj, '__name__'):
        return str(obj.__name__)
    elif _has_own_str_method(obj):
        return str(obj)
    elif hasattr(obj, '__class__'):
        return get_name(obj.__class__)
    else:
        return str(obj)


class HashAlgorithm(Protocol):
    """
    All the hashlib algorithms conform to this, like:
      - hashlib.md5
      - hashlib.sha1
      - hashlib.sha256
      - ...
    """
    def update(self, data: Buffer, /) -> None: ...
    def hexdigest(self) -> str: ...


HashAlgorithmFactory = Callable[[], HashAlgorithm]


def recursive_hash(
    seed: Union[str, None],
    obj: Any,
    alg_factory: HashAlgorithmFactory = lambda: hashlib.sha256(),
) -> str:
    alg = alg_factory()
    def _r(obj: Any) -> None:
        if isinstance(obj, dict):
            items = sorted(obj.items())  # <-- it's critical to sort them so we get canonical hashes
            alg.update('__open_dict__'.encode('utf-8'))
            _r(items)
            alg.update('__close_dict__'.encode('utf-8'))
        elif isinstance(obj, (list, tuple)):
            alg.update('__open_list__'.encode('utf-8'))
            for o in obj:
                _r(o)
            alg.update('__close_list__'.encode('utf-8'))
        elif isinstance(obj, str):
            alg.update('__str__'.encode('utf-8'))
            alg.update(obj.encode('utf-8'))
        elif isinstance(obj, bool):
            alg.update('__bool__'.encode('utf-8'))
            alg.update(f'{1 if obj else 0}'.encode('utf-8'))
        elif isinstance(obj, int):
            alg.update('__int__'.encode('utf-8'))
            alg.update(f'{obj}'.encode('utf-8'))
        elif isinstance(obj, float):
            alg.update('__float__'.encode('utf-8'))
            alg.update(f'{obj:.8e}'.encode('utf-8'))
        elif obj is None:
            alg.update('__None__'.encode('utf-8'))
        else:
            raise ValueError(f"unsupported type: {type(obj)}")
    if seed is not None:
        _r(seed)
    _r(obj)
    return alg.hexdigest()



import pytest

from lasagna.util import (
    parse_docstring,
    combine_pairs,
    convert_to_image_url,
    convert_to_image_base64,
    exponential_backoff_retry_delays,
    get_name,
    recursive_hash,
)

from typing import List

import tempfile
import hashlib
import os


def test_parse_docstring():
    desc, params = parse_docstring("""
        This is the description
        that continues here.
          with indented stuff here
        back to normal.

        :param: x: str: the `x` value

        :param: another: int: another value
    """)
    assert desc == "This is the description\nthat continues here.\n  with indented stuff here\nback to normal."
    assert params == [
        {
            'name': 'x',
            'type': 'str',
            'description': 'the `x` value',
        },
        {
            'name': 'another',
            'type': 'int',
            'description': 'another value',
        },
    ]

    desc, params = parse_docstring("""Hi
        this is the docs.
        :param: x: str: the `x` value
        :param: another: int: another value
    """)
    assert desc == "Hi\nthis is the docs."
    assert params == [
        {
            'name': 'x',
            'type': 'str',
            'description': 'the `x` value',
        },
        {
            'name': 'another',
            'type': 'int',
            'description': 'another value',
        },
    ]

    desc, params = parse_docstring("""
        Hi this is the docs.
        :param: x: str: the `x` value
        :param: another: int: another value
    """)
    assert desc == "Hi this is the docs."
    assert params == [
        {
            'name': 'x',
            'type': 'str',
            'description': 'the `x` value',
        },
        {
            'name': 'another',
            'type': 'int',
            'description': 'another value',
        },
    ]

    desc, params = parse_docstring("""
        Hi this is the docs.

        :param: x: str: the `x` value
        :param: another: int: another value
    """)
    assert desc == "Hi this is the docs."
    assert params == [
        {
            'name': 'x',
            'type': 'str',
            'description': 'the `x` value',
        },
        {
            'name': 'another',
            'type': 'int',
            'description': 'another value',
        },
    ]

    desc, params = parse_docstring("""
        Hi this is the
        docs.


        :param: x: str: the `x` value
        :param: another: int: another value
    """)
    assert desc == "Hi this is the\ndocs."
    assert params == [
        {
            'name': 'x',
            'type': 'str',
            'description': 'the `x` value',
        },
        {
            'name': 'another',
            'type': 'int',
            'description': 'another value',
        },
    ]

    desc, params = parse_docstring("""
        Hi this is the
        docs.


        :param: x: str: the `x` value

        :param: another: int: another value
    """)
    assert desc == "Hi this is the\ndocs."
    assert params == [
        {
            'name': 'x',
            'type': 'str',
            'description': 'the `x` value',
        },
        {
            'name': 'another',
            'type': 'int',
            'description': 'another value',
        },
    ]

    desc, params = parse_docstring("""Hi this is the docs.
        :param: x: str: the `x` value
        :param: another: int: another value
    """)
    assert desc == "Hi this is the docs."
    assert params == [
        {
            'name': 'x',
            'type': 'str',
            'description': 'the `x` value',
        },
        {
            'name': 'another',
            'type': 'int',
            'description': 'another value',
        },
    ]

    desc, params = parse_docstring("""Hi this is the docs.

        :param: x: str: the `x` value
        :param: another: int: another value
    """)
    assert desc == "Hi this is the docs."
    assert params == [
        {
            'name': 'x',
            'type': 'str',
            'description': 'the `x` value',
        },
        {
            'name': 'another',
            'type': 'int',
            'description': 'another value',
        },
    ]

    desc, params = parse_docstring("""
        Hi this is the docs.
        :param: thing: float: (optional) the thing
        :param: another: enum a b other: another value
    """)
    assert desc == "Hi this is the docs."
    assert params == [
        {
            'name': 'thing',
            'type': 'float',
            'description': '(optional) the thing',
            'optional': True,
        },
        {
            'name': 'another',
            'type': 'enum a b other',
            'description': 'another value',
        },
    ]

    desc, params = parse_docstring("""
        Hi this is the docs.
        And this is a second line.
    """)
    assert desc == "Hi this is the docs.\nAnd this is a second line."
    assert params == []

    with pytest.raises(ValueError):
        desc, params = parse_docstring("")

    with pytest.raises(ValueError):
        desc, params = parse_docstring("""
            :param: thing: float: (optional) the thing
            :param: another: enum a b other: another value
        """)

    with pytest.raises(ValueError):
        desc, params = parse_docstring("""

            :param: thing: float: (optional) the thing
            :param: another: enum a b other: another value
        """)

    with pytest.raises(ValueError):
        desc, params = parse_docstring("""
            Docs
            :param: x: badtype: the param
        """)
        print(desc)
        print(params)

    desc, params = parse_docstring("""
        Hi this is the docs.
        :param: thing: float: (optional) the thing
        and a second line for the first param
        :param: another: enum a b other: another value
        and a second line for the second param
    """)
    assert desc == "Hi this is the docs."
    assert params == [
        {
            'name': 'thing',
            'type': 'float',
            'description': '(optional) the thing\nand a second line for the first param',
            'optional': True,
        },
        {
            'name': 'another',
            'type': 'enum a b other',
            'description': 'another value\nand a second line for the second param',
        },
    ]

    desc, params = parse_docstring("""
        Hi this is the
        docs.

        :param: thing: float: (optional) the thing
        and a second line for the first param

        :param: another: enum a b other: another value
        and a second line for the second param
    """)
    assert desc == "Hi this is the\ndocs."
    assert params == [
        {
            'name': 'thing',
            'type': 'float',
            'description': '(optional) the thing\nand a second line for the first param',
            'optional': True,
        },
        {
            'name': 'another',
            'type': 'enum a b other',
            'description': 'another value\nand a second line for the second param',
        },
    ]


def test_combine_pairs():
    def combine_rule_1(a, b):
        return False
    def combine_rule_2(a, b):
        return True, 99
    def combine_rule_3(a, b):
        if a == b:
            return True, a
        return False

    empty: List[int] = []
    oneval: List[int] = [7]
    twovals: List[int] = [7, 12]
    threevals: List[int] = [7, 12, 4]

    assert combine_pairs(empty, combine_rule_1) == []
    assert combine_pairs(empty, combine_rule_2) == []
    assert combine_pairs(empty, combine_rule_3) == []

    assert combine_pairs(oneval, combine_rule_1) == [7]
    assert combine_pairs(oneval, combine_rule_2) == [7]
    assert combine_pairs(oneval, combine_rule_3) == [7]

    assert combine_pairs(twovals, combine_rule_1) == [7, 12]
    assert combine_pairs(twovals, combine_rule_2) == [99]
    assert combine_pairs(twovals, combine_rule_3) == [7, 12]

    assert combine_pairs(threevals, combine_rule_1) == [7, 12, 4]
    assert combine_pairs(threevals, combine_rule_2) == [99, 99]
    assert combine_pairs(threevals, combine_rule_3) == [7, 12, 4]

    assert combine_pairs([1, 4], combine_rule_3) == [1, 4]
    assert combine_pairs([4, 4], combine_rule_3) == [4]

    assert combine_pairs([1, 4, 7], combine_rule_3) == [1, 4, 7]
    assert combine_pairs([4, 4, 7], combine_rule_3) == [4, 7]
    assert combine_pairs([1, 4, 4], combine_rule_3) == [1, 4]
    assert combine_pairs([4, 4, 4], combine_rule_3) == [4, 4]

    assert combine_pairs([1, 4, 7, 9], combine_rule_3) == [1, 4, 7, 9]
    assert combine_pairs([4, 4, 7, 9], combine_rule_3) == [4, 7, 9]
    assert combine_pairs([1, 4, 4, 9], combine_rule_3) == [1, 4, 9]
    assert combine_pairs([1, 4, 7, 7], combine_rule_3) == [1, 4, 7]
    assert combine_pairs([4, 4, 7, 7], combine_rule_3) == [4, 7]
    assert combine_pairs([4, 4, 4, 9], combine_rule_3) == [4, 4, 9]
    assert combine_pairs([1, 7, 7, 7], combine_rule_3) == [1, 7, 7]
    assert combine_pairs([7, 7, 7, 7], combine_rule_3) == [7, 7, 7]


@pytest.mark.asyncio
async def test_convert_to_image_url():
    url = 'https://example.com/img.png'
    s = await convert_to_image_url(url)
    assert s == url

    url = 'http://example.com/img.png'
    s = await convert_to_image_url(url)
    assert s == url

    with tempfile.TemporaryDirectory() as tmp:
        fn = os.path.join(tmp, 'a.jpeg')
        with open(fn, 'wb') as f:
            f.write(b'1234')
        s = await convert_to_image_url(fn)
        assert s == 'data:image/jpeg;base64,MTIzNA=='

        fn = os.path.join(tmp, 'a.png')
        with open(fn, 'wb') as f:
            f.write(b'1235')
        s = await convert_to_image_url(fn)
        assert s == 'data:image/png;base64,MTIzNQ=='

        fn = os.path.abspath(fn)
        with open(fn, 'wb') as f:
            f.write(b'1235')
        s = await convert_to_image_url(fn)
        assert s == 'data:image/png;base64,MTIzNQ=='

        fn_with_schema = f"file://{fn}"
        with open(fn, 'wb') as f:
            f.write(b'1235')
        s = await convert_to_image_url(fn_with_schema)
        assert s == 'data:image/png;base64,MTIzNQ=='

    with pytest.raises(ValueError):
        await convert_to_image_url(os.path.join('does', 'not', 'exist.png'))


@pytest.mark.asyncio
async def test_convert_to_image_base64():
    # Disabled so that our tests don't make remote calls.
    #url = 'https://raw.githubusercontent.com/Rhobota/lasagna-ai/main/logos/lasagna-ai.png'
    #mimetype, s = await convert_to_image_base64(url)
    #assert mimetype == 'image/png'
    #assert len(s) > 1000

    # Disabled so that our tests don't make remote calls.
    #with pytest.raises(Exception):
    #    url = 'https://raw.githubusercontent.com/Rhobota/lasagna-ai/main/logos/DOES_NOT_EXIST.png'
    #    await convert_to_image_base64(url)

    with tempfile.TemporaryDirectory() as tmp:
        fn = os.path.join(tmp, 'a.jpeg')
        with open(fn, 'wb') as f:
            f.write(b'1234')
        mimetype, s = await convert_to_image_base64(fn)
        assert mimetype == 'image/jpeg'
        assert s == 'MTIzNA=='

        fn = os.path.join(tmp, 'a.png')
        with open(fn, 'wb') as f:
            f.write(b'1235')
        mimetype, s = await convert_to_image_base64(fn)
        assert mimetype == 'image/png'
        assert s == 'MTIzNQ=='

        fn = os.path.abspath(fn)
        with open(fn, 'wb') as f:
            f.write(b'1235')
        mimetype, s = await convert_to_image_base64(fn)
        assert mimetype == 'image/png'
        assert s == 'MTIzNQ=='

        fn_with_schema = f"file://{fn}"
        with open(fn, 'wb') as f:
            f.write(b'1235')
        mimetype, s = await convert_to_image_base64(fn_with_schema)
        assert mimetype == 'image/png'
        assert s == 'MTIzNQ=='

    with pytest.raises(ValueError):
        await convert_to_image_base64(os.path.join('does', 'not', 'exist.png'))


def test_exponential_backoff_retry_delays():
    with pytest.raises(AssertionError):
        exponential_backoff_retry_delays(0)
    assert exponential_backoff_retry_delays(1, 3.0, 1e10) == [0.0]
    assert exponential_backoff_retry_delays(2, 3.0, 1e10) == [3.0, 0.0]
    assert exponential_backoff_retry_delays(3, 3.0, 1e10) == [3.0, 9.0, 0.0]
    assert exponential_backoff_retry_delays(4, 3.0, 1e10) == [3.0, 9.0, 27.0, 0.0]
    assert exponential_backoff_retry_delays(4, 5.0, 1e10) == [5.0, 25.0, 125.0, 0.0]
    assert exponential_backoff_retry_delays(4, 5.0, 30.0) == [5.0, 25.0, 30.0, 0.0]


def regular_function():
    pass

def async_regular_function():
    pass

class class_with_str_method:
    def __str__(self) -> str:
        return 'Hi!'

class class_with_no_str_method:
    pass

def test_get_name():
    assert get_name(regular_function) == 'regular_function'
    assert get_name(async_regular_function) == 'async_regular_function'
    assert get_name(class_with_str_method()) == 'Hi!'
    assert get_name(class_with_no_str_method()) == 'class_with_no_str_method'


def test_recursive_hash():
    obj = {
        'b': -5,
        'a': [4, True, False, 'hi', (3.5, 7.0), None],
    }
    got = recursive_hash('hola', obj)
    s = ''.join([v.strip() for v in '''
        __str__hola
        __open_dict__
            __open_list__
                __open_list__
                    __str__a
                    __open_list__
                        __int__4
                        __bool__1
                        __bool__0
                        __str__hi
                        __open_list__
                            __float__3.50000000e+00
                            __float__7.00000000e+00
                        __close_list__
                        __None__
                    __close_list__
                __close_list__
                __open_list__
                    __str__b
                    __int__-5
                __close_list__
            __close_list__
        __close_dict__
        '''.strip().splitlines()])
    correct = hashlib.sha256(s.encode('utf-8')).hexdigest()
    assert got == correct



import pytest

from typing import Callable, List

from lasagna.agent_util import (
    bind_model,
    build_simple_agent,
    build_standard_message_extractor,
    extract_last_message,
    noop_callback,
    partial_bind_model,
    recursive_extract_messages,
    flat_messages,
    build_extraction_agent,
    override_system_prompt,
)

from lasagna.types import (
    AgentCallable,
    BoundAgentCallable,
    AgentRun,
    Message,
    EventPayload,
)

from lasagna.mock_provider import (
    MockProvider,
)

from lasagna.util import get_name

from pydantic import BaseModel, ValidationError


class MyTestType(BaseModel):
    a: str
    b: int


async def _agent_common_test(
    binder: Callable[[AgentCallable], BoundAgentCallable],
    agent: AgentCallable,
):
    events = []
    async def event_callback(event: EventPayload) -> None:
        events.append(event)
    prev_runs: List[AgentRun] = []
    new_run = await binder(agent)(event_callback, prev_runs)
    assert new_run == {
        'agent': 'simple_agent',
        'provider': 'MockProvider',
        'model': 'some_model',
        'model_kwargs': {
            'b': 6,
            'a': 'yes',
        },
        'type': 'messages',
        'messages': [
            {
                'role': 'ai',
                'text': f"model: some_model",
            },
            {
                'role': 'human',
                'text': f"model_kwarg: a = yes",
            },
            {
                'role': 'human',
                'text': f"model_kwarg: b = 6",
            },
        ],
    }
    assert events == [
        ('agent', 'start', 'simple_agent'),
        ('ai', 'text_event', 'Hi!'),
        ('agent', 'end', new_run),
    ]


@pytest.mark.asyncio
async def test_bind_model():
    my_binder = bind_model(MockProvider, 'some_model', a = 'yes', b = 6)
    my_agent = build_simple_agent(name = 'simple_agent')
    await _agent_common_test(my_binder, my_agent)


@pytest.mark.asyncio
async def test_partial_bind_model():
    my_binder = partial_bind_model(MockProvider, 'some_model')(a = 'yes', b = 6)
    my_agent = build_simple_agent(name = 'simple_agent')
    await _agent_common_test(my_binder, my_agent)


@pytest.mark.asyncio
async def test_build_layered_agent():
    my_binder = bind_model(MockProvider, 'a_model')
    my_agent = build_simple_agent(
        name = 'a_layered_agent',
        tools = [],
        message_extractor = build_standard_message_extractor(
            system_prompt_override = 'system test',
        ),
        doc = 'doc test',
    )
    assert my_agent.__doc__ == 'doc test'
    assert str(my_agent) == 'a_layered_agent'
    assert get_name(my_agent) == 'a_layered_agent'
    my_bound_agent = my_binder(my_agent)
    assert my_bound_agent.__doc__ == 'doc test'
    assert my_bound_agent.__name__ == 'a_layered_agent'
    assert get_name(my_bound_agent) == 'a_layered_agent'
    prev_runs: List[AgentRun] = [
        flat_messages(
            'some_agent',
            [
                {
                    'role': 'human',
                    'text': 'layered agent test',
                },
            ],
        ),
    ]
    new_run = await my_bound_agent(noop_callback, prev_runs)
    assert new_run == {
        'agent': 'a_layered_agent',
        'provider': 'MockProvider',
        'model': 'a_model',
        'model_kwargs': {},
        'type': 'messages',
        'messages': [
            {
                'role': 'system',
                'text': 'system test',
            },
            {
                'role': 'human',
                'text': 'layered agent test',
            },
            {
                'role': 'ai',
                'text': f"model: a_model",
            },
        ],
    }


@pytest.mark.asyncio
async def test_model_extract():
    my_binder = bind_model(MockProvider, 'some_model', a = 'yes', b = 6)
    events = []
    async def event_callback(event: EventPayload) -> None:
        events.append(event)
    prev_runs: List[AgentRun] = []
    run = await my_binder(
        build_extraction_agent(
            name = 'my_extraction_agent',
            extraction_type = MyTestType,
        ),
    )(event_callback, prev_runs)
    assert events == [
        (
            'agent',
            'start',
            'my_extraction_agent',
        ),
        (
            'tool_call',
            'tool_call_event',
            {
                'call_id': 'id123',
                'call_type': 'function',
                'function': {
                    'name': 'f',
                    'arguments': '{"a": "yes", "b": 6}',
                },
            },
        ),
        (
            'agent',
            'end',
            {
                'type': 'extraction',
                'messages': [
                    {
                        'role': 'ai',
                        'text': None,
                    },
                ],
                'result': MyTestType(a='yes', b=6),
                'provider': 'MockProvider',
                'agent': 'my_extraction_agent',
                'model': 'some_model',
                'model_kwargs': {
                    'a': 'yes',
                    'b': 6,
                },
            },
        ),
    ]
    assert run['type'] == 'extraction'
    assert run['messages'] == [
        {
            'role': 'ai',
            'text': None,
        },
    ]
    result = run['result']
    assert isinstance(result, MyTestType)
    assert result.a == 'yes'
    assert result.b == 6


@pytest.mark.asyncio
async def test_model_extract_type_mismatch():
    my_binder = bind_model(MockProvider, 'some_model', a = 'yes', b = 'BAD VALUE')
    prev_runs: List[AgentRun] = []
    with pytest.raises(ValidationError):
        await my_binder(
            build_extraction_agent(
                name = 'my_extraction_agent',
                extraction_type = MyTestType,
            ),
        )(noop_callback, prev_runs)


def test_recursive_extract_messages():
    agent_run: AgentRun = {
        'agent': 'outer_agent',
        'type': 'chain',
        'runs': [
            {
                'agent': 'inner_agent_1',
                'type': 'parallel',
                'runs': [
                    {
                        'agent': 'inner_agent_2',
                        'type': 'messages',
                        'messages': [
                            {
                                'role': 'system',
                                'text': 'You are a robot.',
                            },
                            {
                                'role': 'human',
                                'text': 'What are you?',
                            },
                        ],
                    },
                    {
                        'agent': 'inner_agent_3',
                        'type': 'messages',
                        'messages': [
                            {
                                'role': 'system',
                                'text': 'You are a cat.',
                            },
                            {
                                'role': 'human',
                                'text': 'Here kitty kitty!',
                            },
                            {
                                'role': 'tool_res',
                                'tools': [
                                    {
                                        'type': 'any',
                                        'call_id': 'call000',
                                        'result': 'Meow.',
                                    },
                                    {
                                        'type': 'layered_agent',
                                        'call_id': 'call001',
                                        'run': {
                                            'agent': 'inner_agent_4',
                                            'type': 'messages',
                                            'messages': [
                                                {
                                                    'role': 'ai',
                                                    'text': 'Beep.',
                                                },
                                            ],
                                        },
                                    },
                                ],
                            },
                        ],
                    },
                ],
            },
            {
                'agent': 'inner_agent_5',
                'type': 'extraction',
                'messages': [
                    {
                        'role': 'tool_call',
                        'tools': [
                            {
                                'call_id': 'call002',
                                'call_type': 'function',
                                'function': {
                                    'name': 'foo',
                                    'arguments': '{"value": 7}',
                                },
                            },
                        ],
                    },
                ],
                'result': {'value': 7},
            },
            {
                'agent': 'inner_agent_6',
                'type': 'messages',
                'messages': [
                    {
                        'role': 'system',
                        'text': 'You aggregate other AI systems.',
                    },
                    {
                        'role': 'human',
                        'text': 'Summarize the previous AI conversations, please.',
                    },
                ],
            },
        ],
    }
    assert recursive_extract_messages([agent_run], from_layered_agents=True) == [
        {
            'role': 'system',
            'text': 'You are a robot.',
        },
        {
            'role': 'human',
            'text': 'What are you?',
        },
        {
            'role': 'system',
            'text': 'You are a cat.',
        },
        {
            'role': 'human',
            'text': 'Here kitty kitty!',
        },
        {
            'role': 'tool_res',
            'tools': [
                {
                    'type': 'any',
                    'call_id': 'call000',
                    'result': 'Meow.',
                },
                {
                    'type': 'layered_agent',
                    'call_id': 'call001',
                    'run': {
                        'agent': 'inner_agent_4',
                        'type': 'messages',
                        'messages': [
                            {
                                'role': 'ai',
                                'text': 'Beep.',
                            },
                        ],
                    },
                },
            ],
        },
        {
            'role': 'ai',
            'text': 'Beep.',
        },
        {
            'role': 'tool_call',
            'tools': [
                {
                    'call_id': 'call002',
                    'call_type': 'function',
                    'function': {
                        'name': 'foo',
                        'arguments': '{"value": 7}',
                    },
                },
            ],
        },
        {
            'role': 'system',
            'text': 'You aggregate other AI systems.',
        },
        {
            'role': 'human',
            'text': 'Summarize the previous AI conversations, please.',
        },
    ]
    assert recursive_extract_messages([agent_run], from_layered_agents=False) == [
        {
            'role': 'system',
            'text': 'You are a robot.',
        },
        {
            'role': 'human',
            'text': 'What are you?',
        },
        {
            'role': 'system',
            'text': 'You are a cat.',
        },
        {
            'role': 'human',
            'text': 'Here kitty kitty!',
        },
        {
            'role': 'tool_res',
            'tools': [
                {
                    'type': 'any',
                    'call_id': 'call000',
                    'result': 'Meow.',
                },
                {
                    'type': 'layered_agent',
                    'call_id': 'call001',
                    'run': {
                        'agent': 'inner_agent_4',
                        'type': 'messages',
                        'messages': [
                            {
                                'role': 'ai',
                                'text': 'Beep.',
                            },
                        ],
                    },
                },
            ],
        },
        {
            'role': 'tool_call',
            'tools': [
                {
                    'call_id': 'call002',
                    'call_type': 'function',
                    'function': {
                        'name': 'foo',
                        'arguments': '{"value": 7}',
                    },
                },
            ],
        },
        {
            'role': 'system',
            'text': 'You aggregate other AI systems.',
        },
        {
            'role': 'human',
            'text': 'Summarize the previous AI conversations, please.',
        },
    ]
    assert extract_last_message(agent_run, from_layered_agents=True) == {
        'role': 'human',
        'text': 'Summarize the previous AI conversations, please.',
    }


def test_flat_messages():
    messages: List[Message] = [
        {
            'role': 'system',
            'text': 'You are a cat.',
        },
        {
            'role': 'human',
            'text': 'Here kitty kitty!',
        },
    ]
    assert flat_messages('an_agent', messages) == {
        'agent': 'an_agent',
        'type': 'messages',
        'messages': [
            {
                'role': 'system',
                'text': 'You are a cat.',
            },
            {
                'role': 'human',
                'text': 'Here kitty kitty!',
            },
        ],
    }


def test_override_system_prompt():
    messages: List[Message] = []
    assert override_system_prompt(messages, 'Bla') == [
        {
            'role': 'system',
            'text': 'Bla',
        },
    ]
    assert messages == []  # test immutability

    messages: List[Message] = [
        {
            'role': 'system',
            'text': 'You are a robot.',
        },
    ]
    assert override_system_prompt(messages, 'You are Fred.') == [
        {
            'role': 'system',
            'text': 'You are Fred.',
        },
    ]
    assert messages == [   # test immutability
        {
            'role': 'system',
            'text': 'You are a robot.',
        },
    ]

    messages: List[Message] = [
        {
            'role': 'human',
            'text': 'Hi!',
        },
    ]
    assert override_system_prompt(messages, 'You are Fred.') == [
        {
            'role': 'system',
            'text': 'You are Fred.',
        },
        {
            'role': 'human',
            'text': 'Hi!',
        },
    ]
    assert messages == [   # test immutability
        {
            'role': 'human',
            'text': 'Hi!',
        },
    ]

    messages: List[Message] = [
        {
            'role': 'system',
            'text': 'You are a robot.',
        },
        {
            'role': 'human',
            'text': 'Who are you?',
        },
    ]
    assert override_system_prompt(messages, 'You are Fred.') == [
        {
            'role': 'system',
            'text': 'You are Fred.',
        },
        {
            'role': 'human',
            'text': 'Who are you?',
        },
    ]
    assert messages == [   # test immutability
        {
            'role': 'system',
            'text': 'You are a robot.',
        },
        {
            'role': 'human',
            'text': 'Who are you?',
        },
    ]

    messages: List[Message] = [
        {
            'role': 'ai',
            'text': 'Beep.',
        },
        {
            'role': 'human',
            'text': 'Who are you?',
        },
    ]
    assert override_system_prompt(messages, 'You are Fred.') == [
        {
            'role': 'system',
            'text': 'You are Fred.',
        },
        {
            'role': 'ai',
            'text': 'Beep.',
        },
        {
            'role': 'human',
            'text': 'Who are you?',
        },
    ]
    assert messages == [   # test immutability
        {
            'role': 'ai',
            'text': 'Beep.',
        },
        {
            'role': 'human',
            'text': 'Who are you?',
        },
    ]



import pytest

from enum import Enum
from lasagna.pydantic_util import (
    ensure_pydantic_model,
    build_and_validate,
)

from pydantic import BaseModel, ValidationError

from typing import List
from typing_extensions import TypedDict, is_typeddict


class MyPydanticModel(BaseModel):
    a: str
    b: int


class MyColorEnum(Enum):
    RED = 'red'
    GREEN = 'green'
    BLUE = 'blue'


class MySubDict(TypedDict):
    question: str
    answer: str


class MyDict(TypedDict):
    name: str
    age: int
    color: MyColorEnum
    about_me: List[MySubDict]


def test_ensure_pydantic_model_already_is_pydantic():
    assert issubclass(MyPydanticModel, BaseModel)
    m = ensure_pydantic_model(MyPydanticModel)
    assert m is MyPydanticModel
    assert issubclass(m, BaseModel)


def test_ensure_pydantic_model_unknown():
    with pytest.raises(ValueError):
        ensure_pydantic_model(int)


def test_ensure_pydantic_model_typeddict():
    assert is_typeddict(MyDict)
    m = ensure_pydantic_model(MyDict)
    assert not is_typeddict(m)
    assert issubclass(m, BaseModel)

    schema = m.model_json_schema()

    assert schema == {
        "type": "object",
        "title": "MyDict",
        "properties": {
            "name": {
                "title": "Name",
                "type": "string"
            },
            "age": {
                "title": "Age",
                "type": "integer"
            },
            "color": {
                "$ref": "#/$defs/MyColorEnum"
            },
            "about_me": {
                "items": {
                    "$ref": "#/$defs/MySubDict"
                },
                "title": "About Me",
                "type": "array"
            }
        },
        "required": [
            "name",
            "age",
            "color",
            "about_me"
        ],
        "$defs": {
            "MyColorEnum": {
                "enum": [
                    "red",
                    "green",
                    "blue"
                ],
                "title": "MyColorEnum",
                "type": "string"
            },
            "MySubDict": {
                "type": "object",
                "title": "MySubDict",
                "properties": {
                    "question": {
                        "title": "Question",
                        "type": "string"
                    },
                    "answer": {
                        "title": "Answer",
                        "type": "string"
                    }
                },
                "required": [
                    "question",
                    "answer"
                ],
            }
        },
    }


def test_build_and_validate_already_is_pydantic():
    res = build_and_validate(
        MyPydanticModel,
        {
            'a': 'test',
            'b': 7,
        },
    )
    assert res.a == 'test'
    assert res.b == 7
    assert isinstance(res, MyPydanticModel)

    with pytest.raises(ValidationError):
        build_and_validate(
            MyPydanticModel,
            {},   # <-- empty payload!
        )

    with pytest.raises(ValidationError):
        build_and_validate(
            MyPydanticModel,
            {
                'a': 'test',
                #'b': 7,   # <-- missing field!
            },
        )

    with pytest.raises(ValidationError):
        build_and_validate(
            MyPydanticModel,
            {
                'a': 'test',
                'b': 'INVALID',   # <-- invalid value!
            },
        )


def test_build_and_validate_typeddict():
    res = build_and_validate(
        MyDict,
        {
            'name': 'ryan',
            'age': 84,
            'color': 'green',
            'about_me': [
                {
                    'question': 'Why?',
                    'answer': 'because',
                },
                {
                    'question': 'When?',
                    'answer': 'now',
                },
            ],
        },
    )
    assert res['name'] == 'ryan'
    assert res['age'] == 84
    assert res['color'] == MyColorEnum.GREEN
    assert res['about_me'][0]['question'] == 'Why?'
    assert res['about_me'][0]['answer'] == 'because'
    assert res['about_me'][1]['question'] == 'When?'
    assert res['about_me'][1]['answer'] == 'now'
    assert isinstance(res, dict)

    with pytest.raises(ValidationError):
        build_and_validate(
            MyDict,
            {},   # <-- empty payload!
        )

    with pytest.raises(ValidationError):
        build_and_validate(
            MyDict,
            {
                'name': 'ryan',
                'age': 84,
                'color': 'green',
                #'about_me': [],  # <-- missing field!
            },
        )

    with pytest.raises(ValidationError):
        build_and_validate(
            MyDict,
            {
                'name': 'ryan',
                'age': 84,
                'color': 'yellow',  # <-- bad emum!
                'about_me': [],
            },
        )

    with pytest.raises(ValidationError):
        build_and_validate(
            MyDict,
            {
                'name': 'ryan',
                'age': 'INVALID',   # <-- invalid value!
                'color': 'green',
                'about_me': [],
            },
        )

    with pytest.raises(ValidationError):
        build_and_validate(
            MyDict,
            {
                'name': 'ryan',
                'age': 84,
                'color': 'green',
                'about_me': [
                    {
                        'question': 'Why?',
                        #'answer': 'because',  # <-- missing nested field!
                    },
                ],
            },
        )

    with pytest.raises(ValidationError):
        build_and_validate(
            MyDict,
            {
                'name': 'ryan',
                'age': 84,
                'color': 'green',
                'about_me': [
                    {
                        'question': 'Why?',
                        'answer': {
                            'value': 'BAD SCHEMA',   # <-- invalid schema!
                        },
                    },
                ],
            },
        )



from lasagna import (
    known_models,
    build_simple_agent,
)

from lasagna.tui import (
    tui_input_loop,
)

from typing import List, Callable

from dotenv import load_dotenv; load_dotenv()

import asyncio
import aiohttp
import os


MODEL_BINDER = known_models.BIND_OPENAI_gpt_4o_mini()


# LLMs can hallucinate phone numbers, so you should be very cautious
# about letting them have access to this tool without guard rails!
# Below is a whitelist of phone numbers the agent is allowed to send to.
# Thus, if it hallucinates a phone number, we catch its mistake!
PHONE_NUMBER_WHITELIST: List[str] = [
    # PUT YOUR PHONE NUMBER(S) HERE, IN E.164 FORMAT
]


async def send_sms(to_phone_number: str, message_body: str) -> None:
    """
    Use this tool to send SMS message (aka, "text messages").
    :param: to_phone_number: str: the destination phone number to send to (in E.164 format; e.g. '+12223334444', where '+1' is the country code, '222' is the area code and '3334444' is the 7-digit phone number)
    :param: message_body: str: the content (aka, "body") of the SMS message
    """
    if to_phone_number not in PHONE_NUMBER_WHITELIST:
        raise PermissionError('you are not allowed to send SMS messages to that phone number')
    account_sid = os.environ['TWILIO_ACCOUNT_SID']
    account_auth_token = os.environ['TWILIO_AUTH_TOKEN']
    from_phone_number = os.environ['TWILIO_FROM_PHONE']
    payload = {
        'Body': message_body,
        'From': from_phone_number,
        'To':   to_phone_number,
    }
    url = f"https://api.twilio.com/2010-04-01/Accounts/{account_sid}/Messages.json"
    auth = aiohttp.BasicAuth(account_sid, account_auth_token)
    async with aiohttp.ClientSession(raise_for_status=True, auth=auth) as session:
        async with session.post(url, data=payload) as request:
            await request.json()


async def main() -> None:
    system_prompt = "You are grumpy."
    tools: List[Callable] = [
        send_sms,
    ]
    my_agent = build_simple_agent(name = 'agent', tools = tools)
    my_bound_agent = MODEL_BINDER(my_agent)
    await tui_input_loop(my_bound_agent, system_prompt)


if __name__ == '__main__':
    asyncio.run(main())
